#!/usr/bin/env python3
import argparse
import os
import warnings

import matplotlib
import numpy as np
from scipy.cluster.hierarchy import linkage, leaves_list, dendrogram
import pandas as pd
from skbio.diversity.alpha import chao1, shannon, simpson
import dendropy
from joblib import Parallel, delayed
import matplotlib.pyplot as plt
from matplotlib.transforms import Bbox
import matplotlib.gridspec as gridspec
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.ticker import FixedLocator
import seaborn as sns

# from tqdm.auto import tqdm

from masato.get_abundance import get_otu_count, _agg_along_axis, _taxa_qc
from masato.plot_tree import (
    get_taxonomy_tree,
    _calc_y,
    _get_node_label,
    get_coords,
    get_taxon2marker,
    get_taxon2color,
    plot_tree,
)


def _get_dendrogram(df: pd.DataFrame, rows_to_ignore=None):
    """Perform hierarchical clustering on rows of the input dataframe, reorder the rows
    and get the dendrogram, ignoring specified rows.

    Args:
    df (pd.DataFrame): The input dataframe. Rows should be what you want to cluster.
    rows_to_ignore (list, optional): Rows to ignore during clustering. Defaults to None.

    Returns:
    tuple: A tuple containing the reordered dataframe and the linkage matrix.
    """
    if rows_to_ignore is None:
        rows_to_ignore = []
    not_in_df = set(rows_to_ignore) - set(df.index)
    if not_in_df:
        print(
            f"WANRING: `rows_to_ignore` contains rows not in the dataframe: {not_in_df}"
        )
        rows_to_ignore = [i for i in rows_to_ignore if i not in not_in_df]

    # Separate the dataframe into rows to cluster and rows to ignore
    rows_to_cluster = df.index.difference(rows_to_ignore)
    df_to_cluster = df.loc[rows_to_cluster]

    # Perform hierarchical clustering on the rows to be clustered
    linked = linkage(df_to_cluster, "single", optimal_ordering=False)
    order = leaves_list(linked)

    # Reorder the dataframe based on the clustering result, appending ignored rows at the end
    reordered_df = pd.concat(
        [df_to_cluster.iloc[order], df.loc[rows_to_ignore]], axis=0
    )

    return reordered_df, linked


def _stacked_bar(
    dfs: list[pd.DataFrame],
    names: list[str],
    title: str,
    palette: list,
    axs: list[list[Axes]],
    orientation: str = "vertical",
) -> None:
    """Indices of input dataframes are samples, which is non-overlapping. Columns are
    features, which should be exactly the same for all input dataframes.
    """
    for i, (ax_dend, ax, df, name) in enumerate(zip(axs[0], axs[1], dfs, names)):
        if orientation == "horizontal":
            df = df.iloc[::-1]

        if ax_dend is not None:
            df, linked = _get_dendrogram(df)
            dendrogram(
                linked,
                ax=ax_dend,
                leaf_rotation=90,
                orientation="right" if orientation == "horizontal" else "top",
                link_color_func=_black_color_func,
            )
            ax_dend.axis("off")

        if orientation == "vertical":
            df.plot(kind="bar", stacked=True, color=palette, ax=ax)
            ax.set_xlabel("")
            ax.set_ylabel("Relative abundance")
            ax.set_title(name)
            ax.set_ylim(0, 1)
        elif orientation == "horizontal":
            df.plot(kind="barh", stacked=True, color=palette, ax=ax)
            ax.set_ylabel("")
            ax.set_xlabel("Relative abundance")
            ax.set_title(name)
            ax.set_xlim(0, 1)
        else:
            raise ValueError("Unknown orientation.")
        # set xticklabels size
        # ax.tick_params(axis="x", labelsize=6)
        # remove legend unless the last plot
        if i != len(axs[0]) - 1:
            ax.get_legend().remove()
        else:
            if orientation == "horizontal" and ax_dend is not None:
                # put more to the right
                ax.legend(loc="center left", bbox_to_anchor=(1.3, 0.5))
            else:
                ax.legend(loc="center left", bbox_to_anchor=(1.0, 0.5))
            ax.get_legend().set_title(title)
        ax.set_ylabel(ax.yaxis.get_label().get_text(), fontsize=14)
        ax.set_xlabel(ax.xaxis.get_label().get_text(), fontsize=14)


def _black_color_func(*args, **kwargs):
    return "black"


def _heatmap(
    dfs: list[pd.DataFrame],
    names: list[str],
    title: str,
    fig: Figure,
    axs_heatmap: list[Axes],
    axs_dend: list[Axes],
    cmap: str,
    linkage_row: np.ndarray | dendropy.Tree | None = None,
    linkage_row_legend: dict | None = None,
    cbar_label: str | None = None,
    vmax: float | None = None,
    vmin: float | None = None,
    dfs_iso=None,
) -> None:
    dend_row_ax_width = 4 / sum(d.shape[1] for d in dfs)
    if dfs_iso is None:
        dfs_iso = [None] * len(dfs)
    for i, (ax, ax_dend, df, df_iso, name) in enumerate(
        zip(axs_heatmap, axs_dend, dfs, dfs_iso, names, strict=True)
    ):
        # df = df.copy()
        # columns = []
        # for j in df.columns:
        #     if j.endswith("-Swab-combined"):
        #         columns.append(j[1] + "-Bulk")
        #     elif j.endswith("-Scrape-R2A"):
        #         columns.append(j[1] + "-Plate-R2A")
        #     elif j.endswith("-Scrape-TSA"):
        #         columns.append(j[1] + "-Plate-TSA")
        #     else:
        #         raise ValueError(f"Unknown column: {j}")
        # df.columns = columns

        if cbar_label is not None and i == len(axs_heatmap) - 1:
            cbar = True
            # height = np.clip(
            #     axs_heatmap[i - 1].get_position().height * 0.2, 0.3, 0.4
            # )
            # height = 0.4
            height = 5 / df.shape[0]
            cbar_ax_width = 2 / sum(d.shape[1] for d in dfs)
            cbar_ax = fig.add_axes(
                (
                    ax.get_position().x1
                    + (0.04 + (0 if linkage_row is None else dend_row_ax_width)),
                    axs_heatmap[i - 1].get_position().y1 - 1.1 * height,
                    cbar_ax_width,
                    height,
                )
            )
            cbar_ax.yaxis.label.set_size(12)
            cbar_kws = {"label": cbar_label}
        else:
            cbar = False
            cbar_ax = None
            cbar_kws = None

        if ax_dend is not None:
            df, linked = _get_dendrogram(df.transpose())
            df = df.transpose()
        if df_iso is not None:
            df_iso = df_iso.reindex(df.index, axis=0, fill_value="").loc[:, df.columns]
        ax_pos = ax.get_position()
        sns.heatmap(
            df,  # .clip(lower=vmin) if vmin is not None else df,
            ax=ax,
            cmap=cmap,
            lw=0.7,
            linecolor="black",
            cbar=cbar,
            cbar_ax=cbar_ax,
            cbar_kws=cbar_kws,
            vmax=vmax,
            vmin=vmin,
            square=True,
            annot=df_iso,
            fmt="",
        )

        if df_iso is not None:
            sns.heatmap(
                np.zeros_like(df, dtype=float),
                alpha=0,
                ax=ax,
                cmap=None,
                cbar=False,
                square=True,
                xticklabels=df.columns,
                yticklabels=df.index,
                annot=df_iso,
                fmt="",
                annot_kws={"color": "k"},
                mask=~np.isinf(df).to_numpy(),
            )

        # NOTE: I am observing a strange behavior of the heatmap, where the height of
        # the last axis is changed after sns.heatmap while the width is not, which is
        # the opposite of other axes. This is a workaround to revert the height of the
        # last axis after sns.heatmap to before, and infer the correct width based on
        # the per column width from previous axes.
        if i and i == len(axs_heatmap) - 1:
            ax_pos_new = ax.get_position()
            last_ax_pos = axs_heatmap[i - 1].get_position()
            sample_width = (last_ax_pos.x1 - last_ax_pos.x0) / dfs[i - 1].shape[1]
            # keep
            ax.set_position(
                Bbox(
                    np.array(
                        [
                            [ax_pos_new.x0, ax_pos.y0],
                            [ax_pos_new.x0 + sample_width * df.shape[1], ax_pos.y1],
                        ]
                    )
                )
            )

        ax.set_xlabel("")
        ax.tick_params(axis="x", labelrotation=90)

        if ax_dend is not None:
            # enforce ax_dend to have the same width as ax
            ax_dend.set_position(
                (
                    ax.get_position().x0,
                    ax_dend.get_position().y0,
                    ax.get_position().width,
                    ax_dend.get_position().height,
                )
            )
            # plot dendrogram
            dendrogram(
                linked, ax=ax_dend, leaf_rotation=90, link_color_func=_black_color_func
            )
            ax_dend.axis("off")
            ax_dend.set_title(name)
        else:
            ax.set_title(name)

        if linkage_row is not None and i == len(axs_heatmap) - 1:
            dend_row_ax = fig.add_axes(
                (
                    ax.get_position().x1 + 0.005,
                    axs_heatmap[i - 1].get_position().y0,
                    dend_row_ax_width,
                    axs_heatmap[i - 1].get_position().height,
                )
            )
            dend_row_ax.axis("off")
            if isinstance(linkage_row, np.ndarray):
                dendrogram(
                    linkage_row,
                    ax=dend_row_ax,
                    orientation="right",
                    link_color_func=_black_color_func,
                )
                num_elements_in_linkage = linkage_row.shape[0] + 1
                num_elements_not_in_linkage = df.shape[0] - num_elements_in_linkage
                ylow = -10 * num_elements_not_in_linkage
                yhigh = df.shape[0] * 10 + ylow
                dend_row_ax.set_ylim(ylow, yhigh)
            elif isinstance(linkage_row, dendropy.Tree):
                tree = linkage_row
                plot_tree(
                    tree,
                    node_label2marker=tree.taxon2marker,
                    node_label2color=tree.taxon2color,
                    node_label2alpha=tree.taxon2alpha,
                    color_propagate=True,
                    ax=dend_row_ax,
                    nodes_to_drop=tree.nodes_to_drop,
                    terminal_nodes=tree.terminal_nodes,
                )
                yhigh = df.shape[0] - 0.5
                ylow = -0.5
                dend_row_ax.set_ylim(ylow, yhigh)
                dend_row_ax.invert_yaxis()
            else:
                raise ValueError(
                    f"Unknown linkage_row type {type(linkage_row)}. "
                    "Must be np.ndarray or dendropy.Tree."
                )
            if linkage_row_legend is not None:
                has_legend = False
                # add legend with line style
                for label, color in linkage_row_legend.items():
                    if color is None:
                        continue
                    dend_row_ax.plot(
                        [], [], color=color, label=label, linewidth=2, linestyle="-"
                    )
                    has_legend = True
                if has_legend:
                    dend_row_ax.legend(loc="center left", bbox_to_anchor=(1.0, 0.5))

        # turn off y axis ticks and tick labels (set invisible) except for the first panel
        if i == 0:
            ax.set_ylabel(title, fontsize=16)
        else:
            ax.yaxis.set_visible(False)
        # ax.tick_params(axis="x", labelsize=7)


def _barplot_with_whisker_strip(
    dfs: pd.DataFrame,
    names: list[str],
    title: str,
    group_key: str,
    column_of_interest: str,
    plot_strip: bool,
    axs: list[Axes],
    ylog: bool = False,
) -> None:
    axis_tick_fs = 12
    values = []
    for i, (ax, df, name) in enumerate(zip(axs, dfs, names)):
        df = df.copy()
        # index = []
        # for j in df[group_key]:
        #     if j.endswith("-Swab-combined"):
        #         index.append(j[1] + "-Bulk")
        #     elif j.endswith("-Scrape-R2A"):
        #         index.append(j[1] + "-Plate-R2A")
        #     elif j.endswith("-Scrape-TSA"):
        #         index.append(j[1] + "-Plate-TSA")
        #     else:
        #         raise ValueError(f"Unknown column: {j}")
        # df[group_key] = index

        group = df.groupby(group_key, sort=False)
        # Determine which groups have more than one sample
        groups_w_multi_samples = group.filter(lambda x: len(x) > 1)[group_key].unique()

        # assign different colors depending on if group.groups.keys() have "old", "reamplify", or "Scrape"
        # colors = []
        # for key in group.groups.keys():
        #     if key.endswith("-Bulk"):
        #         colors.append("tab:blue")
        #     elif key.endswith("-Plate-R2A"):
        #         colors.append("tab:orange")
        #     elif key.endswith("-Plate-TSA"):
        #         colors.append("tab:green")
        #     else:
        #         raise ValueError(f"Unknown group: {key}")
        colors = "gray"

        # Plot bars with error bars for groups with multiple samples
        value = group[column_of_interest].mean()
        values.extend(value)
        if len(groups_w_multi_samples):
            ax.bar(
                range(len(group)),
                value,
                yerr=group[column_of_interest].std()
                * [
                    1 if group in groups_w_multi_samples else 0
                    for group in group.groups.keys()
                ],
                capsize=2,
                error_kw={"elinewidth": 1.5, "capthick": 1.5},
                width=0.6,
                # fill=False,
                color=colors,
                log=ylog,
            )

            # Remove edgecolor from bars to get rid of the faint gray line
            # for bar in bars:
            #     bar.set_edgecolor("none")

            # Use stripplot for individual points for groups with multiple samples
            if plot_strip:
                df_filtered = df[df[group_key].isin(groups_w_multi_samples)]
                sns.stripplot(
                    x=group_key,
                    y=column_of_interest,
                    data=df_filtered,
                    color="gray",
                    order=group.groups.keys(),
                    ax=ax,
                    size=2,
                )
        else:
            ax.bar(
                range(len(group)),
                value,
                width=0.6,
                # fill=False,
                color=colors,
                log=ylog,
            )
        ax.xaxis.set_major_locator(FixedLocator(range(len(group.groups.keys()))))
        ax.set_xticklabels(
            labels=group.groups.keys(), rotation=90, fontsize=axis_tick_fs
        )
        ax.tick_params(axis="y", labelsize=axis_tick_fs)
        # leave some space on the left and right
        ax.set_xlim(-0.5, len(group.groups.keys()) - 0.5)
        ax.set_title(name)

    axs[0].set_ylabel(title, fontsize=14)
    if ylog:
        # set range to lower than the minimum value and higher than the maximum value
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # possible warnings here:
            # - RuntimeWarning: divide by zero encountered in log10
            # - UserWarning: Attempt to set non-positive ylim on a log-scaled axis with be ignored.
            axs[0].set_ylim(
                10 ** (np.floor(np.log10(min(values)))),
                10 ** (np.ceil(np.log10(max(values)))),
            )


def _calc_alpha_metrics(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate alpha diversity metrics for each sample.
    Note that you should ignore chao1 if your input data is not integer count.
    """
    return pd.DataFrame(
        {
            "chao1": df.apply(chao1, axis=1),
            "richness": df.apply(lambda x: (x > 0).sum(), axis=1),
            "total_counts": df.sum(axis=1),
            "shannon": df.apply(shannon, axis=1),
            "simpson": df.apply(simpson, axis=1),
        },
        index=df.index,
    )


def _get_subplots(
    num_cols: int,
    fig_size: tuple[float, float],
    width_ratios: list[float],
    height_ratios: tuple[float, float] | None = (0.07, 1),
    wspace: float = 0.1,
    hspace: float = 0.01,
    orientation: str = "vertical",
) -> tuple[Figure, list[list[Axes]]]:
    """A helper function to get matplotlib figure and axes for plotting stacked bar plot
    or heatmap with multiple panels.
    """
    if orientation == "horizontal":
        fig_size = fig_size[::-1]
        height_ratios = height_ratios[::-1]
        wspace, hspace = hspace, wspace
        width_ratios, height_ratios = height_ratios, width_ratios

    if height_ratios is None:  # no hierarchical clustering
        if orientation == "vertical":
            fig, axs = plt.subplots(
                1,
                num_cols,
                sharey="row",
                figsize=fig_size,
                width_ratios=width_ratios,
            )
        elif orientation == "horizontal":
            fig, axs = plt.subplots(
                num_cols,
                1,
                sharex="col",
                figsize=fig_size,
                height_ratios=height_ratios,
            )
        else:
            raise ValueError("Unknown orientation.")
        if num_cols == 1:
            axs = [axs]
        # if orientation == "vertical":
        #     for i in range(1, num_cols):
        #         axs[i].sharey(axs[0])
        # elif orientation == "horizontal":
        #     for i in range(1, num_cols):
        #         axs[i].sharex(axs[0])
        axs = [[None] * num_cols, axs]
        fig.subplots_adjust(wspace=wspace, hspace=hspace)
    else:
        if orientation == "vertical":
            if not len(height_ratios) == 2:
                raise ValueError("height_ratios should be a list of length 2.")
            fig, axs = plt.subplots(
                2,
                num_cols,
                figsize=fig_size,
                width_ratios=width_ratios,
                height_ratios=height_ratios,
            )
        elif orientation == "horizontal":
            if not len(width_ratios) == 2:
                raise ValueError("width_ratios should be a list of length 2.")
            fig, axs = plt.subplots(
                num_cols,
                2,
                figsize=fig_size,
                width_ratios=width_ratios,
                height_ratios=height_ratios,
            )
            axs = axs.T
        else:
            raise ValueError("Unknown orientation.")
        if num_cols == 1:
            axs = [[axs[0]], [axs[1]]]
        # share y for the heatmap, i.e., the second row
        if orientation == "vertical":
            for i in range(1, num_cols):
                axs[1][i].sharey(axs[1][0])
        elif orientation == "horizontal":
            axs = axs[::-1]
            for i in range(1, num_cols):
                axs[1][i].sharex(axs[0][1])
        else:
            raise ValueError("Unknown orientation.")
        fig.subplots_adjust(wspace=wspace, hspace=hspace)
    return fig, axs


def _rarefying(
    df: pd.DataFrame, ref: list[int], repeat_num: int = 20
) -> list[pd.DataFrame | list]:
    """Rarefy the dataframe to the reference list or integer.

    The input dataframe should be sample x features, with sample name as index.

    In the returned output, each sample in the input dataframe will be rarefied
    `repeat_num` times, resulting in a dataframe with shape (sample x features x
    repeat_num), except for samples whose reference is themselves, which will be
    repeated only once. Sample name in the returned output will be like
    <original_sample_name>_rarefied_<repeat_num>.
    """
    # make sure all columns in df are positive integers and are in ref.
    if not len(ref) == len(df):
        raise ValueError(
            "The length of ref should be the same as the number of rows in df."
        )

    # get a list of 2d numpy array using joblib parallisim
    res = Parallel(n_jobs=4)(
        delayed(rarefy_array)(df.iloc[idx].to_numpy(), n, repeat_num)
        for idx, n in enumerate(ref)
    )
    sample_names_new_orig = [
        (f"{sample_name}_rarefied_{j}", sample_name)
        for idx, sample_name in enumerate(df.index)
        for j in range(res[idx].shape[0])
    ]
    res = np.concatenate(res, axis=0)
    idx_new, idx_orig = zip(*sample_names_new_orig)
    return pd.DataFrame(res, index=idx_new, columns=df.columns), idx_orig


def rarefy_array(arr: np.ndarray, n: int, k: int, seed: int = 42) -> np.ndarray:
    """Rarefy one row k times so that each row sum up to n."""
    depth = arr.sum()
    rng = np.random.default_rng(seed=seed)
    if n >= depth:  # don't rarefy if expected depth is larger than the actual depth.
        return arr.reshape(1, -1)
    all_elements = np.repeat(np.arange(arr.size), arr)
    return np.stack(
        [
            np.bincount(
                rng.choice(all_elements, size=n, replace=False), minlength=arr.size
            )
            for _ in range(k)
        ],
        axis=0,
    )


def get_abundance_plot(
    df_otu_count: pd.DataFrame,
    df_meta: pd.DataFrame,
    df_tax: pd.DataFrame,
    *,
    df_otu_count_iso: None | pd.DataFrame = None,
    df_meta_iso: None | pd.DataFrame = None,
    df_tax_iso: None | pd.DataFrame = None,
    rep_group_key: str,
    sample_group_key: str,
    tax_levels: list[str],
    rel_ab_thresholds: str,
    plot_type,
    orientation: str = "vertical",
    sample_hierarchical_clustering: bool = False,
    feature_hierarchical_clustering: bool = False,
    feature_ordering,
    isolate_rep_group_key: str,
    isolate_sample_group_key: str,
    keep_rare: bool = True,
    keep_unknown: bool = False,
) -> list[Figure]:
    names, groups = zip(*[i for i in df_meta.groupby(sample_group_key, sort=False)])
    ratio = [i[rep_group_key].nunique() for i in groups]

    if df_otu_count_iso is not None:
        if df_meta_iso is None or df_tax_iso is None:
            raise ValueError(
                "Both metadata and taxonomy tables are required for isolate."
            )
        names_iso, groups_iso = zip(
            *[i for i in df_meta_iso.groupby(sample_group_key, sort=False)]
        )
        if not df_tax.equals(df_tax_iso):
            raise ValueError("Taxonomy tables are not the same.")
    else:
        df_otu_count_iso = df_meta_iso = df_tax_iso = None

    if plot_type == "heatmap_raw":
        df_otu_rel_ab_g = _agg_along_axis(
            df_otu_count,
            df_meta[sample_group_key] + "\t" + df_meta[rep_group_key],
            axis=0,
            aggfunc="sum",
        )
    else:
        # process into relative abundance and aggregate at replication group level
        df_otu_rel_ab_g = _agg_along_axis(
            df_otu_count.div(df_otu_count.sum(axis=1), axis=0),
            df_meta[sample_group_key] + "\t" + df_meta[rep_group_key],
            axis=0,
        )

    if df_otu_count_iso is not None:
        df_otu_count_iso_g = _agg_along_axis(
            df_otu_count_iso,
            df_meta_iso[isolate_sample_group_key or sample_group_key]
            + "\t"
            + df_meta_iso[isolate_rep_group_key or rep_group_key],
            axis=0,
            aggfunc="sum",
        )
        # drop None index
        df_otu_count_iso_g = df_otu_count_iso_g.loc[~df_otu_count_iso_g.index.isna()]

    if len(rel_ab_thresholds) == 1:
        rel_ab_thresholds = rel_ab_thresholds * len(tax_levels)

    # an empirical way to determine width for pretty figure
    width = df_otu_rel_ab_g.shape[0] / 3.5 + 1.5
    figs = []
    for level, rel_ab_thres in zip(tax_levels, rel_ab_thresholds):
        # aggregate at taxonomic level
        res = _taxa_qc(
            _agg_along_axis(df_otu_rel_ab_g, df_tax[level], axis=1),
            rel_ab_thres,
            keep_rare=keep_rare,
            keep_unknown=keep_unknown,
        )
        sorted_taxa = sorted(res.columns)
        res = res[sorted_taxa]

        if df_otu_count_iso is not None:
            res_iso = _taxa_qc(
                _agg_along_axis(df_otu_count_iso_g, df_tax[level], axis=1),
                0,
                keep_rare=keep_rare,
                keep_unknown=keep_unknown,
            )
            other_taxa = set(res_iso.columns) - set(res.columns)
            if other_taxa:
                # aggregate those into an "other" columns
                res_iso["others"] = res_iso[list(other_taxa)].sum(axis=1)
                res_iso = res_iso.drop(other_taxa, axis=1)
                # drop others if all zero
                if not res_iso["others"].sum():
                    res_iso = res_iso.drop("others", axis=1)
            bulk_samples = set(res.index)
            iso_samples = set(res_iso.index)
            if not bulk_samples == iso_samples:
                # take the intersect and print warning for removing samples
                rep_group_both = bulk_samples & iso_samples
                if not rep_group_both:
                    raise ValueError("No common replication group.")
                bulk_removed = bulk_samples - rep_group_both
                iso_removed = iso_samples - rep_group_both
                print("WARNING: Samples in bulk and isolate are not the same. ")
                if bulk_removed:
                    print(f"Samples removed from bulk: {bulk_removed}.")
                if iso_removed:
                    print(f"Samples removed from isolate: {iso_removed}.")
                # Take the union of the two by filling in zero while respecting the order of the bulk

                # res = res[lambda x: x in rep_group_both]
                # res_iso = res_iso[lambda x: x in rep_group_both]
                res = res.query("index in @rep_group_both")
                res_iso = res_iso.query("index in @rep_group_both")
                width = res.shape[0] / 4 + 1.5

            res_iso = res_iso[[i for i in sorted_taxa if i in res_iso.columns]]
            res_iso_group_list = [
                res_iso.loc[
                    pd.Index(
                        name
                        + "\t"
                        + group[isolate_rep_group_key or rep_group_key].unique()
                    ).intersection(res_iso.index)
                ]
                for name, group in zip(names_iso, groups_iso)
            ]
            for res_iso_group in res_iso_group_list:
                res_iso_group.index = res_iso_group.index.map(
                    lambda x: x.split("\t")[1]
                )
        else:
            res_iso_group_list = None

        res_group_list = [
            res.loc[
                pd.Index(name + "\t" + group[rep_group_key].unique()).intersection(
                    res.index
                )
            ]
            for name, group in zip(names, groups)
        ]
        # fix index
        for res_group in res_group_list:
            res_group.index = res_group.index.map(lambda x: x.split("\t")[1])

        num_cols = len(res_group_list)
        # wspace, hspace = 0.1, 0.01

        fig, axs_dend, axs = None, None, None
        if plot_type in ["stacked_bar", "all"]:
            # stacked bar plot
            fig_size = (width, 4)
            custom_palette = (
                sns.color_palette("tab20", 20)
                + sns.color_palette("tab20b", 20)
                + sns.color_palette("tab20c", 20)
            )
            fig, axs = _get_subplots(
                num_cols,
                fig_size,
                ratio,
                height_ratios=None,
                orientation=orientation,
            )
            _stacked_bar(
                res_group_list,
                names,
                f"Taxonomy at {level if level != 'otu' else level.upper()} level",
                custom_palette,
                axs,
                orientation=orientation,
            )

        if plot_type in [
            "heatmap",
            "heatmap_log10",
            "heatmap_binary",
            "heatmap_raw",
            "all",
        ]:
            size = (
                res.shape[0] // (res.shape[0] / width),
                max(1, res.shape[1] // (res.shape[0] / width)),
            )
            if plot_type in ["heatmap_log10", "heatmap", "all"]:
                if sample_hierarchical_clustering:
                    fig, (axs_dend, axs) = _get_subplots(num_cols, size, ratio)
                else:
                    fig, (axs_dend, axs) = _get_subplots(
                        num_cols, size, ratio, height_ratios=None
                    )
                # pesudo_abundance = 1e-4
                vmax, vmin = 1, 1e-4

                def _get_log10(arr: pd.DataFrame) -> pd.DataFrame:
                    # take log10, change -inf to log10(vmin), take transpose
                    arr = arr.copy()
                    arr = np.log10(arr)
                    # arr[arr == 0] = vmin
                    return arr

                if plot_type == "heatmap":
                    dfs = [
                        res_group.transpose().replace(0, np.nan)
                        for res_group in res_group_list
                    ]
                elif plot_type in ["heatmap_log10", "all"]:
                    # suppress warning for 0 in log calculation
                    with np.errstate(divide="ignore"):
                        dfs = [
                            _get_log10(res_group.transpose())
                            for res_group in res_group_list
                        ]
                else:
                    raise ValueError(f"Unknown plot type: {plot_type}")
                if df_otu_count_iso is not None:

                    def num_picks_to_symbol(num_picks: int) -> str:
                        if num_picks == 0:
                            return ""
                        elif num_picks == 1:
                            return "+"
                        elif num_picks < 10:
                            return "⁎"  # six pointed black star
                        elif num_picks < 50:
                            return "⁎⁎"
                        elif num_picks < 100:
                            return "⁂"
                        else:
                            return "⁎⁎\n**"

                    dfs_iso = [
                        res_iso_group.transpose().map(num_picks_to_symbol)
                        for res_iso_group in res_iso_group_list
                    ]
                else:
                    dfs_iso = None
                df = pd.concat(dfs, axis=1)
                taxon2color = None
                if feature_ordering == "hierarchical":
                    if plot_type == "heatmap_log10":
                        min_val = df.replace(-np.inf, np.nan).min().min()
                        df = df.replace(-np.inf, min_val)
                    else:
                        df = df.fillna(0)
                    df, linkage_row = _get_dendrogram(
                        df, rows_to_ignore=["unknown", "others"]
                    )
                    dfs_temp = []
                    i = 0
                    for df_temp in dfs:
                        dfs_temp.append(df.iloc[:, i : i + df_temp.shape[1]])
                        i += df_temp.shape[1]
                    dfs = dfs_temp
                elif feature_ordering == "taxonomy_tree":
                    df_tax_filtered = df_tax.query(f"{level}.isin(@dfs[0].index)")
                    tree = get_taxonomy_tree(df_tax_filtered)
                    taxon2marker = get_taxon2marker(df_tax_filtered)
                    ordered_taxon = _calc_y(
                        tree,
                        df_tax_filtered,
                        collapse_level=level,
                        # base=dfs[0].shape[0] - len(tree.leaf_nodes()),
                    )
                    get_coords(
                        tree.seed_node,
                        max_x=tree.seed_node.distance_from_tip()
                        + tree.seed_node.distance_from_root(),
                    )
                    # df_tax_filtered["temp_sort_by"] = pd.Categorical(
                    #     df_tax_filtered[level], categories=ordered_taxon, ordered=True
                    # )
                    idxs = np.unique(ordered_taxon, return_index=True)[1]
                    ordered_taxon = [ordered_taxon[i] for i in sorted(idxs)]
                    df_tax_filtered = df_tax_filtered.iloc[
                        pd.Categorical(
                            df_tax_filtered[level],
                            categories=ordered_taxon,
                            ordered=True,
                        ).argsort()
                    ]

                    taxon2color = get_taxon2color(
                        df_tax_filtered,
                        levels_of_interest=["order", "phylum"],
                        target_level=level,
                    )
                    taxon2alpha = {
                        k: v for k, v in df_tax_filtered.genus_p.to_dict().items()
                    }
                    tree.taxon2marker = taxon2marker
                    tree.taxon2color = taxon2color
                    tree.taxon2alpha = taxon2alpha
                    tree.nodes_to_drop = set(df_tax.index)
                    tree.terminal_nodes = set(ordered_taxon)
                    # reorder the dataframe rows (taxa) to match the tree
                    taxa_other = dfs[0].index.difference(ordered_taxon).tolist()
                    dfs = [df.loc[ordered_taxon + taxa_other] for df in dfs]
                    linkage_row = tree
                else:
                    linkage_row = None
                _heatmap(
                    dfs,
                    names,
                    title=f"Taxonomy at {level} level",
                    cbar_label=(
                        f"log10(relative abundance)\n(range: [{vmin:.0e}, {vmax}])"
                        if plot_type == "heatmap_log10"
                        else "Relative abundance"
                    ),
                    fig=fig,
                    axs_heatmap=axs,
                    axs_dend=axs_dend,
                    cmap="rocket_r",
                    linkage_row=linkage_row,
                    linkage_row_legend=taxon2color,
                    vmax=np.log10(vmax) if plot_type == "heatmap_log10" else vmax,
                    vmin=np.log10(vmin) if plot_type == "heatmap_log10" else vmin,
                    dfs_iso=dfs_iso,
                )
            if plot_type in ["heatmap_binary"]:
                # size = size[0], size[1] / 2
                if sample_hierarchical_clustering:
                    fig, (axs_dend, axs) = _get_subplots(num_cols, size, ratio)
                    # axs_heatmap, axs_dend = axs
                else:
                    fig, (axs_dend, axs) = _get_subplots(
                        num_cols, size, ratio, height_ratios=None
                    )
                    # axs_heatmap = axs
                    axs_dend = [None] * num_cols
                dfs = [(res_group > 0).astype(int).T for res_group in res_group_list]
                df = pd.concat(dfs, axis=1)
                if feature_hierarchical_clustering:
                    df, linkage_row = _get_dendrogram(
                        df, rows_to_ignore=["unknown", "others"]
                    )
                    dfs_temp = []
                    i = 0
                    for df_temp in dfs:
                        dfs_temp.append(df.iloc[:, i : i + df_temp.shape[1]])
                        i += df_temp.shape[1]
                    dfs = dfs_temp
                else:
                    linkage_row = None
                _heatmap(
                    dfs,
                    names,
                    title=f"Taxonomy at {level} level",
                    fig=fig,
                    axs_heatmap=axs,
                    axs_dend=axs_dend,
                    cmap="BuPu",
                    linkage_row=linkage_row,
                )
            if plot_type in ["heatmap_raw"]:
                if sample_hierarchical_clustering:
                    fig, (axs_dend, axs) = _get_subplots(num_cols, size, ratio)
                else:
                    fig, (axs_dend, axs) = _get_subplots(
                        num_cols, size, ratio, height_ratios=None
                    )
                vmax, vmin = 1e3, 0
                with np.errstate(divide="ignore"):
                    dfs = [
                        np.log10(res_group).transpose() for res_group in res_group_list
                    ]
                df = pd.concat(dfs, axis=1)
                if feature_ordering == "hierarchical":
                    df, linkage_row = _get_dendrogram(
                        df, rows_to_ignore=["unknown", "others"]
                    )
                    dfs_temp = []
                    i = 0
                    for df_temp in dfs:
                        dfs_temp.append(df.iloc[:, i : i + df_temp.shape[1]])
                        i += df_temp.shape[1]
                    dfs = dfs_temp
                elif feature_ordering == "taxonomy_tree":
                    df_tax_filtered = df_tax.query(f"{level}.isin(@dfs[0].index)")
                    tree = get_taxonomy_tree(df_tax_filtered)
                    taxon2marker = get_taxon2marker(df_tax_filtered)
                    ordered_taxon = _calc_y(
                        tree,
                        df_tax_filtered,
                        collapse_level=level,
                        # base=dfs[0].shape[0] - len(tree.leaf_nodes()),
                    )
                    get_coords(
                        tree.seed_node,
                        max_x=tree.seed_node.distance_from_tip()
                        + tree.seed_node.distance_from_root(),
                    )
                    # df_tax_filtered["temp_sort_by"] = pd.Categorical(
                    #     df_tax_filtered[level], categories=ordered_taxon, ordered=True
                    # )
                    idxs = np.unique(ordered_taxon, return_index=True)[1]
                    ordered_taxon = [ordered_taxon[i] for i in sorted(idxs)]
                    df_tax_filtered = df_tax_filtered.iloc[
                        pd.Categorical(
                            df_tax_filtered[level],
                            categories=ordered_taxon,
                            ordered=True,
                        ).argsort()
                    ]

                    taxon2color = get_taxon2color(
                        df_tax_filtered,
                        levels_of_interest=["order", "phylum"],
                        target_level=level,
                    )
                    taxon2alpha = {
                        k: v for k, v in df_tax_filtered.genus_p.to_dict().items()
                    }
                    tree.taxon2marker = taxon2marker
                    tree.taxon2color = taxon2color
                    tree.taxon2alpha = taxon2alpha
                    tree.nodes_to_drop = set(df_tax.index)
                    tree.terminal_nodes = set(ordered_taxon)
                    # reorder the dataframe rows (taxa) to match the tree
                    taxa_other = dfs[0].index.difference(ordered_taxon).tolist()
                    dfs = [df.loc[ordered_taxon + taxa_other] for df in dfs]
                    linkage_row = tree
                else:
                    linkage_row = None
                _heatmap(
                    dfs,
                    names,
                    title=f"Taxonomy at {level} level",
                    cbar_label=f"log10(#picked isolates)\n(range: [{vmin}, {vmax:.0e}])",
                    # cbar_label="Count",
                    fig=fig,
                    axs_heatmap=axs,
                    axs_dend=axs_dend,
                    # cmap=sns.cubehelix_palette(as_cmap=True),
                    cmap="summer",
                    linkage_row=linkage_row,
                    linkage_row_legend=taxon2color,
                    vmax=np.log10(vmax),
                    vmin=0,
                )
        if fig is None:
            raise ValueError(f"No figure is created because of invalid plot type {plot_type}.")
        figs.append(fig)
    return figs


def main():
    # if __name__ == "__main__":

    # matplotlib.use("TkAgg")
    plt.rcParams["pdf.fonttype"] = 42
    plt.rcParams["svg.fonttype"] = "none"

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="command")

    parser_ab = subparsers.add_parser("abundance_group")
    parser_ab.add_argument(
        "plot_type",
        type=str,
        choices=[
            "stacked_bar",
            "heatmap",
            "heatmap_log10",
            "heatmap_binary",
            "heatmap_raw",
            "all",
        ],
    )
    parser_ab.add_argument("-i", "--otu_count_tsv", type=str, required=True)
    parser_ab.add_argument("-m", "--metadata", type=str, required=True)
    parser_ab.add_argument("-t", "--otu_taxonomy_tsv", type=str, required=True)
    parser_ab.add_argument("-f", "--fig_dir", type=str, required=True)
    parser_ab.add_argument("-r", "--rep_group_key", type=str, default="rep_group")
    parser_ab.add_argument("-s", "--sample_group_key", type=str, default="sample_group")
    parser_ab.add_argument("-sp", "--spikein_taxa_key", type=str, default="spike_in")
    parser_ab.add_argument(
        "-l", "--tax_levels", nargs="+", default=["order", "family", "genus", "otu"]
    )
    parser_ab.add_argument(
        "-a",
        "--rel_ab_thresholds",
        nargs="+",
        type=float,
        default=[0.0],
        help="Genus relative abundance threshold.",
    )
    parser_ab.add_argument(
        "-ori",
        "--orientation",
        type=str,
        choices=["vertical", "horizontal"],
        default="vertical",
    )
    parser_ab.add_argument(
        "-sc",
        "--sample_hierarchical_clustering",
        action="store_true",
        default=False,
        help="Perform hierarchical clustering on samples in each group, order them "
        "accordingly and add a dendrogram at the top. Only affects heatmap, not "
        "stacked bar plot.",
    )
    parser_ab.add_argument(
        "-fo",
        "--feature_ordering",
        type=str,
        choices=["alphabetical", "hierarchical", "taxonomy_tree"],
        default="alphabetical",
        help="Perform hierarchical clustering on features in each group, order them "
        "accordingly and add a dendrogram at the left. Only affects heatmap, not "
        "stacked bar plot.",
    )
    parser_ab.add_argument("-ii", "--isolate_otu_count_tsv", type=str, default=None)
    parser_ab.add_argument("-im", "--isolate_metadata", type=str, default=None)
    parser_ab.add_argument(
        "-ir",
        "--isolate_rep_group_key",
        type=str,
        default=None,
    )
    parser_ab.add_argument(
        "-is",
        "--isolate_sample_group_key",
        type=str,
        default=None,
    )
    parser_ab.add_argument(
        "-isp",
        "--isolate_spikein_taxa_key",
        type=str,
        default=None,
    )

    parser_stats = subparsers.add_parser("stats_sample_count")
    parser_stats.add_argument("-i", "--otu_count_tsv", type=str)
    parser_stats.add_argument("-m", "--metadata", type=str)
    parser_stats.add_argument("-t", "--otu_taxonomy_tsv", type=str)
    parser_stats.add_argument("-f", "--fig_dir", type=str)
    parser_stats.add_argument(
        "-s", "--sample_group_key", type=str, default="sample_group"
    )
    parser_stats.add_argument("-r", "--rep_group_key", type=str, default="rep_group")
    parser_stats.add_argument("-sp", "--spikein_taxa_key", type=str, default="spike_in")
    parser_stats.add_argument(
        "-l", "--tax_levels", nargs="+", default=["order", "family", "genus", "otu"]
    )
    parser_stats.add_argument("-rk", "--rarefying_key", type=str, default=None)
    parser_stats.add_argument("-rv", "--rarefying_value", type=int, default=None)
    parser_stats.add_argument("-rn", "--rarefying_repeat", type=int, default=0)
    parser_stats.add_argument("-p", "--plot_strip", action="store_true", default=False)

    args = parser.parse_args()

    otu_count_tsv = args.otu_count_tsv
    metadata_tsv = args.metadata
    otu_taxonomy_tsv = args.otu_taxonomy_tsv
    fig_dir = args.fig_dir
    rep_group_key = args.rep_group_key
    sample_group_key = args.sample_group_key
    spikein_taxa_key = args.spikein_taxa_key
    tax_levels = args.tax_levels

    # ====== read tables and preprocess ======
    df_otu_count, df_meta, df_tax = get_otu_count(
        otu_count_tsv,
        metadata_tsv,
        otu_taxonomy_tsv,
        sample_weight_key=None,
        spikein_taxa_key=spikein_taxa_key,
    )

    names, groups = zip(*[i for i in df_meta.groupby(sample_group_key, sort=False)])
    ratio = [i[rep_group_key].nunique() for i in groups]
    os.makedirs(fig_dir, exist_ok=True)

    if args.command == "abundance_group":
        rel_ab_thresholds = args.rel_ab_thresholds
        plot_type = args.plot_type
        orientation = args.orientation
        sample_hierarchical_clustering = args.sample_hierarchical_clustering
        # feature_hierarchical_clustering = args.feature_hierarchical_clustering
        feature_ordering = args.feature_ordering
        isolate_otu_count_tsv = args.isolate_otu_count_tsv
        isolate_metadata_tsv = args.isolate_metadata
        isolate_rep_group_key = args.isolate_rep_group_key
        isolate_sample_group_key = args.isolate_sample_group_key
        isolate_spikein_taxa_key = args.isolate_spikein_taxa_key
        # no_normalize = args.no_normalize

        if isolate_otu_count_tsv is not None:
            df_otu_count_iso, df_meta_iso, df_tax_iso = get_otu_count(
                isolate_otu_count_tsv,
                isolate_metadata_tsv,
                otu_taxonomy_tsv,
                sample_weight_key=None,
                spikein_taxa_key=isolate_spikein_taxa_key or spikein_taxa_key,
                warning=False,
            )
            names_iso, groups_iso = zip(
                *[i for i in df_meta_iso.groupby(sample_group_key, sort=False)]
            )
            if not df_tax.equals(df_tax_iso):
                raise ValueError("Taxonomy tables are not the same.")
        else:
            df_otu_count_iso = df_meta_iso = df_tax_iso = None

        if plot_type == "heatmap_raw":
            df_otu_rel_ab_g = _agg_along_axis(
                df_otu_count,
                df_meta[sample_group_key] + "\t" + df_meta[rep_group_key],
                axis=0,
                aggfunc="sum",
            )
        else:
            # process into relative abundance and aggregate at replication group level
            df_otu_rel_ab_g = _agg_along_axis(
                df_otu_count.div(df_otu_count.sum(axis=1), axis=0),
                df_meta[sample_group_key] + "\t" + df_meta[rep_group_key],
                axis=0,
            )

        if df_otu_count_iso is not None:
            df_otu_count_iso_g = _agg_along_axis(
                df_otu_count_iso,
                df_meta_iso[isolate_sample_group_key or sample_group_key]
                + "\t"
                + df_meta_iso[isolate_rep_group_key or rep_group_key],
                axis=0,
                aggfunc="sum",
            )
            # drop None index
            df_otu_count_iso_g = df_otu_count_iso_g.loc[
                ~df_otu_count_iso_g.index.isna()
            ]

        if len(rel_ab_thresholds) == 1:
            rel_ab_thresholds = rel_ab_thresholds * len(tax_levels)

        # an empirical way to determine width for pretty figure
        width = df_otu_rel_ab_g.shape[0] / 3.5 + 1.5
        for level, rel_ab_thres in zip(tax_levels, rel_ab_thresholds):
            # aggregate at taxonomic level
            res = _taxa_qc(
                _agg_along_axis(df_otu_rel_ab_g, df_tax[level], axis=1),
                rel_ab_thres,
                keep_rare=True,
                keep_unknown=False,
            )
            sorted_taxa = sorted(res.columns)
            res = res[sorted_taxa]

            if df_otu_count_iso is not None:
                res_iso = _taxa_qc(
                    _agg_along_axis(df_otu_count_iso_g, df_tax[level], axis=1),
                    0,
                    keep_rare=True,
                    keep_unknown=False,
                )
                other_taxa = set(res_iso.columns) - set(res.columns)
                if other_taxa:
                    # aggregate those into an "other" columns
                    res_iso["others"] = res_iso[list(other_taxa)].sum(axis=1)
                    res_iso = res_iso.drop(other_taxa, axis=1)
                bulk_samples = set(res.index)
                iso_samples = set(res_iso.index)
                if not bulk_samples == iso_samples:
                    # take the intersect and print warning for removing samples
                    rep_group_both = bulk_samples & iso_samples
                    if not rep_group_both:
                        raise ValueError("No common replication group.")
                    bulk_removed = bulk_samples - rep_group_both
                    iso_removed = iso_samples - rep_group_both
                    print("WARNING: Samples in bulk and isolate are not the same. ")
                    if bulk_removed:
                        print(f"Samples removed from bulk: {bulk_removed}.")
                    if iso_removed:
                        print(f"Samples removed from isolate: {iso_removed}.")
                    # Take the union of the two by filling in zero while respecting the order of the bulk

                    # res = res[lambda x: x in rep_group_both]
                    # res_iso = res_iso[lambda x: x in rep_group_both]
                    res = res.query("index in @rep_group_both")
                    res_iso = res_iso.query("index in @rep_group_both")
                    width = res.shape[0] / 4 + 1.5

                res_iso = res_iso[[i for i in sorted_taxa if i in res_iso.columns]]
                res_iso_group_list = [
                    res_iso.loc[
                        pd.Index(
                            name
                            + "\t"
                            + group[isolate_rep_group_key or rep_group_key].unique()
                        ).intersection(res_iso.index)
                    ]
                    for name, group in zip(names_iso, groups_iso)
                ]
                for res_iso_group in res_iso_group_list:
                    res_iso_group.index = res_iso_group.index.map(
                        lambda x: x.split("\t")[1]
                    )
            else:
                res_iso_group_list = None

            res_group_list = [
                res.loc[
                    pd.Index(name + "\t" + group[rep_group_key].unique()).intersection(
                        res.index
                    )
                ]
                for name, group in zip(names, groups)
            ]
            # fix index
            for res_group in res_group_list:
                res_group.index = res_group.index.map(lambda x: x.split("\t")[1])

            num_cols = len(res_group_list)
            # wspace, hspace = 0.1, 0.01

            if plot_type in ["stacked_bar", "all"]:
                # stacked bar plot
                fig_size = (width, 4)
                custom_palette = (
                    sns.color_palette("tab20", 20)
                    + sns.color_palette("tab20b", 20)
                    + sns.color_palette("tab20c", 20)
                )
                fig, axs = _get_subplots(
                    num_cols,
                    fig_size,
                    ratio,
                    height_ratios=None,
                    orientation=orientation,
                )
                _stacked_bar(
                    res_group_list,
                    names,
                    f"Taxonomy at {level if level != 'otu' else level.upper()} level",
                    custom_palette,
                    axs,
                    orientation=orientation,
                )
                fig.savefig(
                    f"{fig_dir}/rel_ab_group_{level}_sb.png",
                    bbox_inches="tight",
                    dpi=300,
                )
                fig.savefig(
                    f"{fig_dir}/rel_ab_group_{level}_sb.svg",
                    bbox_inches="tight",
                    dpi=300,
                )
            if plot_type in [
                "heatmap",
                "heatmap_log10",
                "heatmap_binary",
                "heatmap_raw",
                "all",
            ]:
                size = (
                    res.shape[0] // (res.shape[0] / width),
                    max(1, res.shape[1] // (res.shape[0] / width)),
                )
                if plot_type in ["heatmap_log10", "heatmap", "all"]:
                    if sample_hierarchical_clustering:
                        fig, (axs_dend, axs) = _get_subplots(num_cols, size, ratio)
                    else:
                        fig, (axs_dend, axs) = _get_subplots(
                            num_cols, size, ratio, height_ratios=None
                        )
                    # pesudo_abundance = 1e-4
                    vmax, vmin = 1, 1e-4

                    def _get_log10(arr: pd.DataFrame) -> pd.DataFrame:
                        # take log10, change -inf to log10(vmin), take transpose
                        arr = arr.copy()
                        arr = np.log10(arr)
                        # arr[arr == 0] = vmin
                        return arr

                    if plot_type == "heatmap":
                        dfs = [
                            res_group.transpose().replace(0, np.nan)
                            for res_group in res_group_list
                        ]
                    elif plot_type in ["heatmap_log10", "all"]:
                        # suppress warning for 0 in log calculation
                        with np.errstate(divide="ignore"):
                            dfs = [
                                _get_log10(res_group.transpose())
                                for res_group in res_group_list
                            ]
                    else:
                        raise ValueError(f"Unknown plot type: {plot_type}")
                    if df_otu_count_iso is not None:

                        def num_picks_to_symbol(num_picks: int) -> str:
                            if num_picks == 0:
                                return ""
                            elif num_picks == 1:
                                return "+"
                            elif num_picks < 10:
                                return "⁎"  # six pointed black star
                            elif num_picks < 50:
                                return "⁎⁎"
                            elif num_picks < 100:
                                return "⁂"
                            else:
                                return "⁎⁎\n**"

                        dfs_iso = [
                            res_iso_group.transpose().map(num_picks_to_symbol)
                            for res_iso_group in res_iso_group_list
                        ]
                    else:
                        dfs_iso = None
                    df = pd.concat(dfs, axis=1)
                    taxon2color = None
                    if feature_ordering == "hierarchical":
                        if plot_type == "heatmap_log10":
                            min_val = df.replace(-np.inf, np.nan).min().min()
                            df = df.replace(-np.inf, min_val)
                        else:
                            df = df.fillna(0)
                        df, linkage_row = _get_dendrogram(
                            df, rows_to_ignore=["unknown", "others"]
                        )
                        dfs_temp = []
                        i = 0
                        for df_temp in dfs:
                            dfs_temp.append(df.iloc[:, i : i + df_temp.shape[1]])
                            i += df_temp.shape[1]
                        dfs = dfs_temp
                    elif feature_ordering == "taxonomy_tree":
                        df_tax_filtered = df_tax.query(f"{level}.isin(@dfs[0].index)")
                        tree = get_taxonomy_tree(df_tax_filtered)
                        taxon2marker = get_taxon2marker(df_tax_filtered)
                        ordered_taxon = _calc_y(
                            tree,
                            df_tax_filtered,
                            collapse_level=level,
                            # base=dfs[0].shape[0] - len(tree.leaf_nodes()),
                        )
                        get_coords(
                            tree.seed_node,
                            max_x=tree.seed_node.distance_from_tip()
                            + tree.seed_node.distance_from_root(),
                        )
                        # df_tax_filtered["temp_sort_by"] = pd.Categorical(
                        #     df_tax_filtered[level], categories=ordered_taxon, ordered=True
                        # )
                        idxs = np.unique(ordered_taxon, return_index=True)[1]
                        ordered_taxon = [ordered_taxon[i] for i in sorted(idxs)]
                        df_tax_filtered = df_tax_filtered.iloc[
                            pd.Categorical(
                                df_tax_filtered[level],
                                categories=ordered_taxon,
                                ordered=True,
                            ).argsort()
                        ]

                        taxon2color = get_taxon2color(
                            df_tax_filtered,
                            levels_of_interest=["order", "phylum"],
                            target_level=level,
                        )
                        taxon2alpha = {
                            k: v for k, v in df_tax_filtered.genus_p.to_dict().items()
                        }
                        tree.taxon2marker = taxon2marker
                        tree.taxon2color = taxon2color
                        tree.taxon2alpha = taxon2alpha
                        tree.nodes_to_drop = set(df_tax.index)
                        tree.terminal_nodes = set(ordered_taxon)
                        # reorder the dataframe rows (taxa) to match the tree
                        taxa_other = dfs[0].index.difference(ordered_taxon).tolist()
                        dfs = [df.loc[ordered_taxon + taxa_other] for df in dfs]
                        linkage_row = tree
                    else:
                        linkage_row = None
                    _heatmap(
                        dfs,
                        names,
                        title=f"Taxonomy at {level} level",
                        cbar_label=(
                            f"log10(relative abundance)\n(range: [{vmin:.0e}, {vmax}])"
                            if plot_type == "heatmap_log10"
                            else "Relative abundance"
                        ),
                        fig=fig,
                        axs_heatmap=axs,
                        axs_dend=axs_dend,
                        cmap="rocket_r",
                        linkage_row=linkage_row,
                        linkage_row_legend=taxon2color,
                        vmax=np.log10(vmax) if plot_type == "heatmap_log10" else vmax,
                        vmin=np.log10(vmin) if plot_type == "heatmap_log10" else vmin,
                        dfs_iso=dfs_iso,
                    )
                    fig.savefig(
                        f"{fig_dir}/rel_ab_group_{level}_hm{'l' if plot_type == 'heatmap_log10' else ''}.png",
                        bbox_inches="tight",
                        dpi=300,
                    )
                    fig.savefig(
                        f"{fig_dir}/rel_ab_group_{level}_hm{'l' if plot_type == 'heatmap_log10' else ''}.svg",
                        bbox_inches="tight",
                        dpi=300,
                    )
                # if plot_type in ["heatmap"]:
                #     if sample_hierarchical_clustering:
                #         fig, (axs_dend, axs) = _get_subplots(num_cols, size, ratio)
                #     else:
                #         fig, (axs_dend, axs) = _get_subplots(
                #             num_cols, size, ratio, height_ratios=None
                #         )
                #     dfs = ([res_group.T for res_group in res_group_list],)
                #     df = pd.concat(dfs, axis=1)
                #     if feature_ordering == "hierarchical":
                #         df, linkage_row = _get_dendrogram(
                #             df, rows_to_ignore=["unknown", "others"]
                #         )
                #         dfs_temp = []
                #         i = 0
                #         for df_temp in dfs:
                #             dfs_temp.append(df.iloc[:, i : i + df_temp.shape[1]])
                #             i += df_temp.shape[1]
                #         dfs = dfs_temp
                #     else:
                #         linkage_row = None
                #     _heatmap(
                #         dfs,
                #         names,
                #         title=f"Taxonomy at {level} level",
                #         cbar_label="relative abundance",
                #         fig=fig,
                #         axs_heatmap=axs,
                #         axs_dend=axs_dend,
                #         cmap=None,
                #         linkage_row=linkage_row,
                #     )
                #     fig.savefig(
                #         f"{fig_dir}/rel_ab_group_{level}_hm.png",
                #         bbox_inches="tight",
                #         dpi=300,
                #     )
                #     fig.savefig(
                #         f"{fig_dir}/rel_ab_group_{level}_hm.svg",
                #         bbox_inches="tight",
                #         dpi=300,
                #     )
                if plot_type in ["heatmap_binary"]:
                    # size = size[0], size[1] / 2
                    if sample_hierarchical_clustering:
                        fig, (axs_dend, axs) = _get_subplots(num_cols, size, ratio)
                        # axs_heatmap, axs_dend = axs
                    else:
                        fig, (axs_dend, axs) = _get_subplots(
                            num_cols, size, ratio, height_ratios=None
                        )
                        # axs_heatmap = axs
                        axs_dend = [None] * num_cols
                    dfs = [
                        (res_group > 0).astype(int).T for res_group in res_group_list
                    ]
                    df = pd.concat(dfs, axis=1)
                    if feature_hierarchical_clustering:
                        df, linkage_row = _get_dendrogram(
                            df, rows_to_ignore=["unknown", "others"]
                        )
                        dfs_temp = []
                        i = 0
                        for df_temp in dfs:
                            dfs_temp.append(df.iloc[:, i : i + df_temp.shape[1]])
                            i += df_temp.shape[1]
                        dfs = dfs_temp
                    else:
                        linkage_row = None
                    _heatmap(
                        dfs,
                        names,
                        title=f"Taxonomy at {level} level",
                        fig=fig,
                        axs_heatmap=axs,
                        axs_dend=axs_dend,
                        cmap="BuPu",
                        linkage_row=linkage_row,
                    )
                    fig.savefig(
                        f"{fig_dir}/rel_ab_group_{level}_hmb.png",
                        bbox_inches="tight",
                        dpi=300,
                    )
                    fig.savefig(
                        f"{fig_dir}/rel_ab_group_{level}_hmb.svg",
                        bbox_inches="tight",
                        dpi=300,
                    )
                if plot_type in ["heatmap_raw"]:
                    if sample_hierarchical_clustering:
                        fig, (axs_dend, axs) = _get_subplots(num_cols, size, ratio)
                    else:
                        fig, (axs_dend, axs) = _get_subplots(
                            num_cols, size, ratio, height_ratios=None
                        )
                    vmax, vmin = 1e3, 0
                    with np.errstate(divide="ignore"):
                        dfs = [
                            np.log10(res_group).transpose()
                            for res_group in res_group_list
                        ]
                    df = pd.concat(dfs, axis=1)
                    if feature_ordering == "hierarchical":
                        df, linkage_row = _get_dendrogram(
                            df, rows_to_ignore=["unknown", "others"]
                        )
                        dfs_temp = []
                        i = 0
                        for df_temp in dfs:
                            dfs_temp.append(df.iloc[:, i : i + df_temp.shape[1]])
                            i += df_temp.shape[1]
                        dfs = dfs_temp
                    elif feature_ordering == "taxonomy_tree":
                        df_tax_filtered = df_tax.query(f"{level}.isin(@dfs[0].index)")
                        tree = get_taxonomy_tree(df_tax_filtered)
                        taxon2marker = get_taxon2marker(df_tax_filtered)
                        ordered_taxon = _calc_y(
                            tree,
                            df_tax_filtered,
                            collapse_level=level,
                            # base=dfs[0].shape[0] - len(tree.leaf_nodes()),
                        )
                        get_coords(
                            tree.seed_node,
                            max_x=tree.seed_node.distance_from_tip()
                            + tree.seed_node.distance_from_root(),
                        )
                        # df_tax_filtered["temp_sort_by"] = pd.Categorical(
                        #     df_tax_filtered[level], categories=ordered_taxon, ordered=True
                        # )
                        idxs = np.unique(ordered_taxon, return_index=True)[1]
                        ordered_taxon = [ordered_taxon[i] for i in sorted(idxs)]
                        df_tax_filtered = df_tax_filtered.iloc[
                            pd.Categorical(
                                df_tax_filtered[level],
                                categories=ordered_taxon,
                                ordered=True,
                            ).argsort()
                        ]

                        taxon2color = get_taxon2color(
                            df_tax_filtered,
                            levels_of_interest=["order", "phylum"],
                            target_level=level,
                        )
                        taxon2alpha = {
                            k: v for k, v in df_tax_filtered.genus_p.to_dict().items()
                        }
                        tree.taxon2marker = taxon2marker
                        tree.taxon2color = taxon2color
                        tree.taxon2alpha = taxon2alpha
                        tree.nodes_to_drop = set(df_tax.index)
                        tree.terminal_nodes = set(ordered_taxon)
                        # reorder the dataframe rows (taxa) to match the tree
                        taxa_other = dfs[0].index.difference(ordered_taxon).tolist()
                        dfs = [df.loc[ordered_taxon + taxa_other] for df in dfs]
                        linkage_row = tree
                    else:
                        linkage_row = None
                    _heatmap(
                        dfs,
                        names,
                        title=f"Taxonomy at {level} level",
                        cbar_label=f"log10(#picked isolates)\n(range: [{vmin}, {vmax:.0e}])",
                        # cbar_label="Count",
                        fig=fig,
                        axs_heatmap=axs,
                        axs_dend=axs_dend,
                        # cmap=sns.cubehelix_palette(as_cmap=True),
                        cmap="summer",
                        linkage_row=linkage_row,
                        linkage_row_legend=taxon2color,
                        vmax=np.log10(vmax),
                        vmin=0,
                    )
                    fig.savefig(
                        f"{fig_dir}/rel_ab_group_{level}_hmr.png",
                        bbox_inches="tight",
                        dpi=300,
                    )
                    fig.savefig(
                        f"{fig_dir}/rel_ab_group_{level}_hmr.svg",
                        bbox_inches="tight",
                        dpi=300,
                    )
    elif args.command == "stats_sample_count":
        plot_strip = args.plot_strip
        rarefying_key = args.rarefying_key
        rarefying_value = args.rarefying_value
        rarefying_repeat = args.rarefying_repeat

        width = df_meta.groupby([sample_group_key, rep_group_key]).ngroups / 3 + 1.5

        if rarefying_repeat > 1:
            # Rarefying takes place by the following order:
            # - When rarefying_key is specified:
            #    - When this key is a string, rarefy to the depth of the sample corresponding to the key.
            #    - When this key is a int, rarefy to this value.
            #    - When this key is None, rarefy to `rarefying_value` if it is
            #        specified, otherwise rarefy to the minimum depth of all samples.
            # - When rarefying_key isn't specified, treat as if rarefying_key is None
            #     for all samples.
            depth = df_otu_count.sum(axis=1)
            if rarefying_value is None:
                rarefying_value = int(depth.min()) - 1

            ref = []
            if rarefying_key is None:
                rarefying_keys = [np.nan] * len(df_otu_count)
            else:
                rarefying_keys = df_meta[rarefying_key].tolist()

            for i in rarefying_keys:
                if isinstance(i, str):
                    if not i in df_otu_count.index:
                        raise ValueError(f"Reference sample {i} is not available.")
                    ref.append(depth[i])
                elif isinstance(i, int):
                    ref.append(i)
                elif np.isnan(i):
                    ref.append(rarefying_value)
                else:
                    raise ValueError(
                        f"Unknown data type for rarefying_key: {i} ({type(i)})"
                    )

            # rarefy
            df_otu_count_orig = df_otu_count.copy()
            df_meta_orig = df_meta.copy()
            df_otu_count, names_orig = _rarefying(df_otu_count, ref, rarefying_repeat)
            # NOTE:
            # Must create the dummy dataframe as a column, cannot be empty dataframe
            # with index, otherwise order of merged dataframe index will be slightly
            # wrong.
            df_meta = pd.merge(
                pd.DataFrame({"original_sample_name": names_orig}),
                df_meta,
                left_on="original_sample_name",
                right_index=True,
                validate="many_to_one",
            ).reset_index(names=df_meta.index.name)
            df_meta.index = df_otu_count.index
        else:
            df_otu_count_orig = None
            df_meta_orig = None

        wspace = 0.15
        for level in tax_levels:
            # no need for taxa qc, all taxa count
            res = _agg_along_axis(df_otu_count, df_tax[level], axis=1)
            meta_l = pd.merge(
                df_meta,
                _calc_alpha_metrics(res),
                left_index=True,
                right_index=True,
            )
            _, groups = zip(*[i for i in meta_l.groupby(sample_group_key, sort=False)])
            title_prefix = f"{'ZOTU' if level == 'otu' else level.capitalize()}"
            for column_of_interest, title in zip(
                ["shannon", "simpson", "richness", "chao1", "total_counts"],
                [
                    "Shannon entropy",
                    "Simpson's index",
                    "Richness",
                    "Chao1 index",
                    "Total counts",
                ],
            ):
                title = f"{title_prefix} {title}"
                fig, axs = plt.subplots(
                    1, len(groups), sharey="row", width_ratios=ratio, figsize=(width, 3)
                )
                if len(groups) == 1:
                    axs = [axs]
                _barplot_with_whisker_strip(
                    groups,
                    names=names,
                    title=title,
                    group_key=rep_group_key,
                    column_of_interest=column_of_interest,
                    plot_strip=plot_strip,
                    axs=axs,
                    ylog=False,
                )
                fig.subplots_adjust(wspace=wspace)
                for format_ in ["png", "svg"]:
                    fig.savefig(
                        f"{fig_dir}/{column_of_interest}_{level}.{format_}",
                        bbox_inches="tight",
                        dpi=300,
                    )
                # close figure to save memory
                plt.close(fig)

        # plot sequencing depth separately
        if df_otu_count_orig is not None:
            df_otu_count = df_otu_count_orig
            df_meta = df_meta_orig
        res = df_otu_count  # no need to aggregate
        meta_l = pd.merge(
            df_meta,
            _calc_alpha_metrics(res),
            left_index=True,
            right_index=True,
        )
        _, groups = zip(*[i for i in meta_l.groupby(sample_group_key, sort=False)])
        fig, axs = plt.subplots(
            1, len(groups), sharey="row", width_ratios=ratio, figsize=(width, 3)
        )
        if len(groups) == 1:
            axs = [axs]
        fig.subplots_adjust(wspace=wspace)
        _barplot_with_whisker_strip(
            groups,
            names=names,
            title="Sequencing depth",
            group_key=rep_group_key,
            column_of_interest="sequencing_depth",
            plot_strip=plot_strip,
            axs=axs,
            ylog=True,
        )
        for format_ in ["png", "svg"]:
            fig.savefig(
                f"{fig_dir}/sequencing_depth.{format_}",
                bbox_inches="tight",
                dpi=300,
            )
        plt.close(fig)
    else:
        raise ValueError("Unknown command.")



import pytest
from pathlib import Path


@pytest.fixture
def test_data_dir():
    return Path(__file__).parent / "data"


def test_get_tree(script_runner, test_data_dir):
    """
    $src_dir/get_tree.py -i $working_dir/$output_unoise3_dir/unoise3_zotu.fa
    """
    # Prepare input and output directories
    input_file = test_data_dir / "output_unoise3_16s/unoise3_zotu.fa"
    output_file = test_data_dir / "output_unoise3_16s/unoise3_zotu.newick"

    # Run the script
    ret = script_runner.run(["get_tree.py", "-i", str(input_file)])

    # Assertions
    assert ret.success, f"Script failed with return code {ret.returncode}"
    assert output_file.exists(), f"Output file not created: {output_file}"
    # Optionally, add more assertions to verify the content of the output file



#!/usr/bin/env python

import glob
import os
import shutil
import subprocess
import argparse
from pathlib import Path

from Bio.Seq import Seq
from matplotlib.pyplot import bar

from masato.utils import cat_fastq, cat_fastq_se, smart_open, print_command


def get_rc(seq: str) -> str:
    """Return the reverse complement of the given sequence."""
    return str(Seq(seq).reverse_complement())


PRIMER_ITS_5 = "GTACCTGCGGARGGATCA"
PRIMER_ITS_7 = "ATGAGATCCRTTGYTRAAAGTT"
PRIMER_16S_5 = "GTGTGYCAGCMGCCGCGGTAA"
PRIMER_16S_7 = "CCGGACTACNVGGGTWTCTAAT"
PRIMER_16S_V3V4_341F = "CCTACGGGNGGCWGCAG"
PRIMER_16S_V3V4_805R = "GACTACHVGGGTATCTAATCC"
PRIMER_TN5 = "AGATGTGTATAAGAGACAG"
# PRIMER_MAPS_0 = "GCGAGACACTGCTGAGCG"
# PRIMER_MAPS_1 = "GTCACTCCAGCCTCGTCG"
# PRIMER_MAPS_2 = "AGGCAGCGTCTGTAGCGA"
PRIMER_MAPS_1 = "GCGAGACACTGCTGAGCG"
PRIMER_MAPS_2 = "GTCACTCCAGCCTCGTCG"
PRIMER_MAPS_3 = "GTGTTCGTCGGCAGCGTC" + PRIMER_TN5
PRIMER_MAPS_r = "GTCTCGTGGGCTCGG" + PRIMER_TN5
ECREC_SPACER0 = "GAGCACAAATATCATCGCTCAAACC"
ECREC_DR = "GTGTTCCCCGCGCCAGCGGGGATAAACC"
ECREC_LEADER = "CTGGCTTAAAAAATCATTAATTAATAATAGGTTATGTTTAGA"
RECORDING_PRIMER_3 = "AGATCGGAAGAGCACACGTCTGA"
RECORDING_PRIMER_5 = "CCTACACGACGCTCTTCCGATCT"


def run_trim_galore(input_dir, output_dir, pair):
    """Run trim_galore on the given pair of files."""
    r1_file, r2_file = pair
    cmd = [
        "trim_galore",
        "--paired",
        os.path.join(input_dir, r1_file),
        os.path.join(input_dir, r2_file),
        "-a",
        "GGCTGACTGACT",
        "-o",
        output_dir,
        "-j",
        "4",
    ]
    return subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


def run_trimmomatic(input_fastq_1: str, input_fastq_2: str, output_dir: str) -> None:
    """Run trimmomatic on the given pair of files using subprocess, following this example:
    trimmomatic-0.30.jar PE s_1_1_sequence.txt.gz s_1_2_sequence.txt.gz
    lane1_forward_paired.fq.gz lane1_forward_unpaired.fq.gz lane1_reverse_paired.fq.gz
    lane1_reverse_unpaired.fq.gz ILLUMINACLIP:TruSeq3-PE.fa:2:30:10 LEADING:3
    TRAILING:3 SLIDINGWINDOW:4:15 MINLEN:36
    """
    output_fastq_1 = os.path.join(output_dir, os.path.basename(input_fastq_1))
    output_fastq_2 = os.path.join(output_dir, os.path.basename(input_fastq_2))
    output_fastq_1_unpaired = output_fastq_1.replace(
        ".fastq", ".trimmomatic.unpaired.fastq"
    )
    output_fastq_2_unpaired = output_fastq_2.replace(
        ".fastq", ".trimmomatic.unpaired.fastq"
    )
    output_trim_log = os.path.join(output_dir, "trimmomatic.log")
    output_summary = os.path.join(output_dir, "trimmomatic.summary")
    os.makedirs(output_dir, exist_ok=True)
    subprocess.run(
        [
            "trimmomatic",
            "PE",
            "-threads",
            "8",
            "-trimlog",
            output_trim_log,
            "-summary",
            output_summary,
            input_fastq_1,
            input_fastq_2,
            output_fastq_1,
            output_fastq_1_unpaired,
            output_fastq_2,
            output_fastq_2_unpaired,
            # "ILLUMINACLIP:TruSeq3-PE.fa:2:30:10",
            "LEADING:3",
            "TRAILING:3",
            "SLIDINGWINDOW:4:15",
            "MINLEN:36",
        ]
    )


def run_fastp(
    input_fastq_1: str, input_fastq_2: str, output_dir: str, stdout: bool = False
) -> subprocess.Popen[bytes]:
    output_fastq_1 = os.path.join(output_dir, os.path.basename(input_fastq_1))
    output_fastq_2 = os.path.join(output_dir, os.path.basename(input_fastq_2))
    output_fastq_1_unpaired = output_fastq_1.replace(".fastq", ".fastp.unpaired.fastq")
    output_fastq_2_unpaired = output_fastq_2.replace(".fastq", ".fastp.unpaired.fastq")
    failed_out = output_fastq_2.replace(".fastq", ".fastp.failed.fastq")
    report_out = os.path.join(output_dir, "fastp.report.json")
    os.makedirs(output_dir, exist_ok=True)
    fastp_proc = subprocess.Popen(
        [
            "fastp",
            "-i",
            input_fastq_1,
            "-I",
            input_fastq_2,
            "--unpaired1",
            output_fastq_1_unpaired,
            "--unpaired2",
            output_fastq_2_unpaired,
            "--failed_out",
            failed_out,
            # cut 5'
            "-5",
            "--cut_front_window_size",
            "1",
            "--cut_front_mean_quality",
            "3",
            # cut 3'
            "-3",
            "--cut_tail_window_size",
            "1",
            "--cut_tail_mean_quality",
            "3",
            # cut right
            "-r",
            "--cut_right_window_size",
            "4",
            "--cut_right_mean_quality",
            "15",
            "--length_required",
            "36",
            "--json",
            report_out,
            "-w",
            "8",
        ]
        + (
            [
                "-o",
                output_fastq_1,
                "-O",
                output_fastq_2,
            ]
            if stdout is False
            else ["--stdout"]
        ),
        stdout=None if stdout is False else subprocess.PIPE,
    )
    return fastp_proc


def run_fastp_cutadapt(
    input_fastq_1: str,
    input_fastq_2: str,
    output_dir: str,
    amplicon_type: str = "ITS",
):
    """Quality trimming and filtering by fastp, then adapter trimming by cutadapt.
    For read 1, trim RC of i7 adapter from the 3' end; For read 2, trim RC of i5 adapter
        also from the 3' end. Just trim and save report, no need to separate trimmed
        and untrimmed reads or discard reads.
    """
    output_fastq_1: str = os.path.join(output_dir, os.path.basename(input_fastq_1))
    output_fastq_2: str = os.path.join(output_dir, os.path.basename(input_fastq_2))
    report_report: str = os.path.join(output_dir, "cutadapt.report.json")
    if amplicon_type not in ["ITS", "16S"]:
        raise ValueError("amplicon_type must be either ITS or 16S.")

    fastp_proc = run_fastp(input_fastq_1, input_fastq_2, output_dir, stdout=True)
    _ = subprocess.run(
        [
            "cutadapt",
            "-a",
            get_rc(PRIMER_ITS_7 if amplicon_type == "ITS" else PRIMER_16S_7),
            "-A",
            get_rc(PRIMER_ITS_5 if amplicon_type == "ITS" else PRIMER_16S_5),
            "-o",
            output_fastq_1,
            "-p",
            output_fastq_2,
            "--json",
            report_report,
            "--interleaved",
            "-",
        ],
        stdin=fastp_proc.stdout,
    )
    fastp_proc.wait()


def simple_preprocess(fastq_dir: str, output_fastq: str) -> None:
    """Given a directory of fastq files, perform renaming, quality trimming, adapter
    trimming and length filtering using just python and fastp. fastp takes interleaved
    input from stdin, to which we write the renamed and interleaved paired-end reads.

    Note:
    - Adapter sequences are optional since fastp can trim paired-end reads by overlap
        analysis, which is more versatile and robust.
    - fastp can also merge pairs, but we don't do it here but with the following ZOTU
        pipeline, just to be more coherent.
    """
    output_dir = os.path.dirname(output_fastq)
    os.makedirs(output_dir, exist_ok=True)
    os.makedirs(os.path.join(output_dir, "fastp"), exist_ok=True)
    fastp_proc = subprocess.Popen(
        [
            "fastp",
            # "-i",
            # os.path.join(fastq_dir, "*_R1_001.fastq.gz"),
            # "-I",
            # os.path.join(fastq_dir, "*_R2_001.fastq.gz"),
            "--stdin",
            "--interleaved_in",
            output_fastq,
            "-w",  # number of threads
            "8",
            # don't trim, so that amplicons have uniform length, as suggested by Edgar
            "--length_required",
            "200",
            # if you don't want to merge, modify the following arguments
            # for the meaning of each of the following arguments, see fastp github README
            "--merge",
            "--merged_out",
            output_fastq,
            "--unpaired1",
            os.path.join(output_dir, "fastp", "unpaired_R1.fastq.gz"),
            "--unpaired2",
            os.path.join(output_dir, "fastp", "unpaired_R2.fastq.gz"),
            "--failed_out",
            os.path.join(output_dir, "fastp", "failed.fastq.gz"),
            "--out1",
            os.path.join(output_dir, "fastp", "unmerged_R1.fastq.gz"),
            "--out2",
            os.path.join(output_dir, "fastp", "unmerged_R2.fastq.gz"),
            "--html",  # report
            os.path.join(output_dir, "fastp", "report.html"),
            "--json",
            os.path.join(output_dir, "fastp", "report.json"),
        ],
        stdin=subprocess.PIPE,
        # stdout=subprocess.PIPE,
    )
    cat_fastq(fastq_dir, output_fp_r1=fastp_proc.stdin, output_fp_r2=fastp_proc.stdin)
    fastp_proc.stdin.close()
    fastp_proc.wait()


def pseudo_merge(
    fastq_dir: str,
    output_fastq: str,
    threads: int,
    min_l1: int,
    first_k1: int,
    min_l2: int,
    first_k2: int,
    cutadapt_error: float = 0.1,
) -> None:
    """
    Process all FASTQ files in a directory by performing quality trimming, adapter trimming,
    and merging paired-end reads.

    Args:
        fastq_dir (str): Directory containing FASTQ files.
        output_dir (str): Directory to save processed FASTQ files.
        threads (int): Number of threads to use.
        min_l1 (int): Minimum length for R1 reads after trimming.
        first_k1 (int): Target length for R1 reads after trimming.
        min_l2 (int): Minimum length for R2 reads after trimming.
        first_k2 (int): Target length for R2 reads after trimming.
        cutadapt_error (float): Error tolerance for cutadapt trimming.
    """
    output_dir = os.path.dirname(output_fastq)
    os.makedirs(output_dir, exist_ok=True)
    log_dir = Path(output_dir) / "trim_logs"
    os.makedirs(log_dir, exist_ok=True)

    fastp_html_r1 = log_dir / "fastp.html"
    fastp_json_r1 = log_dir / "fastp.json"

    trimmed_r1 = Path(output_dir) / "trimmed_1.fastq.gz"
    trimmed_r2 = Path(output_dir) / "trimmed_2.fastq.gz"

    # Process R1 using cat_fastq_se() directly into fastp.
    # Unpaired or failed reads are discarded.
    fastp_args = [
        "fastp",
        "--stdin",
        "--interleaved_in",
        "--stdout",
        "--cut_right",
        "--html",
        str(fastp_html_r1),
        "--json",
        str(fastp_json_r1),
        "--thread",
        str(threads),
    ]
    print_command(fastp_args)
    fastp_proc = subprocess.Popen(
        fastp_args,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    )

    # too short reads are discarded.
    cutadapt_args = [
        "cutadapt",
        "--interleaved",
        "--minimum-length",
        f"{min_l1}:{min_l2}",
        "--length",
        f"{first_k1}",
        "-L",  # there is not a long version of this argument, but it is -l for R2. This is a 4.9 feature.
        f"{first_k2}",
        "--pair-filter",
        "both",
        "-e",
        str(cutadapt_error),
        "-o",
        str(trimmed_r1),
        "-p",
        "-",
        "--cores",
        str(threads),
        "-",
    ]
    print_command(cutadapt_args)
    cutadapt_proc = subprocess.Popen(
        cutadapt_args,
        stdin=fastp_proc.stdout,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    )

    # fastp_min_length_args = [
    #     "fastp",
    #     "--stdin",
    #     "--interleaved_in",
    #     "--out1",
    #     str(trimmed_r1),
    #     "--out2",  # out2 to stdout
    #     "/dev/stdout",
    #     "--max_len1",
    #     str(first_k1),
    #     "--max_len2",
    #     str(first_k2),
    #     "--thread",
    #     str(threads),
    # ]
    # print_command(fastp_min_length_args)
    # fastp_min_length_proc = subprocess.Popen(
    #     fastp_min_length_args,
    #     stdin=cutadapt_proc.stdout,
    #     stdout=subprocess.PIPE,
    #     stderr=subprocess.DEVNULL,
    # )

    seqkit_args = ["seqkit", "seq", "-rp", "-t", "DNA", "-o", str(trimmed_r2)]
    print_command(seqkit_args)
    seqkit_proc = subprocess.Popen(
        seqkit_args,
        stdin=cutadapt_proc.stdout,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    # Pipe R1 reads into fastp
    cat_fastq(fastq_dir, output_fp_r1=fastp_proc.stdin)

    fastp_proc.stdin.close()
    fastp_proc.stdout.close()
    cutadapt_proc.stdout.close()
    seqkit_proc.wait()
    # wait for all processes to finish
    # fastp_proc.wait()
    # cutadapt_proc.wait()
    # fastp_min_length_proc.wait()
    # seqkit_proc.wait()

    # Concatenate trimmed R1 and R2
    _merge_fastq_gz(str(trimmed_r1), str(trimmed_r2), output_fastq)
    # os.remove(trimmed_r1)
    # os.remove(trimmed_r2)


def rename_files_with_mmv(file_dir: str, patterns_file: str) -> None:
    # Step 1: Resolve absolute path of the patterns file
    patterns_file_abs = os.path.abspath(patterns_file)

    # Save the current working directory
    original_dir = os.getcwd()

    # Step 2: Change to the temporary directory
    os.chdir(file_dir)

    # Step 3: Execute the mmv command
    # with open(patterns_file_abs) as file:
    #     subprocess.run(["mmv", "-d"], stdin=file)
    with open(patterns_file_abs) as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            old_name, new_name = line.split()
            subprocess.run(["mv", old_name, new_name])

    # Step 4: Change back to the original directory
    os.chdir(original_dir)


def get_min_overlap(adapter: str, frac: float = 0.8) -> int:
    return int(len(adapter) * frac)


def get_primer_set(name: str) -> tuple[str, str | None] | dict[str, str] | str:
    if name == "its":
        return (
            f"^{PRIMER_ITS_5};required...{get_rc(PRIMER_ITS_7)};optional",
            f"^{PRIMER_ITS_7};required...{get_rc(PRIMER_ITS_5)};optional",
        )
    elif name == "16s":
        return (
            f"^{PRIMER_16S_5};required...{get_rc(PRIMER_16S_7)};optional",
            f"^{PRIMER_16S_7};required...{get_rc(PRIMER_16S_5)};optional",
        )
    elif name == "its_3":
        return get_rc(PRIMER_ITS_7), get_rc(PRIMER_ITS_5)
    elif name == "16s_3":
        return get_rc(PRIMER_16S_7), get_rc(PRIMER_16S_5)
    # elif name == "maps_0":
    #     # optional
    #     return f"{PRIMER_MAPS_0}...{get_rc(PRIMER_MAPS_r)}", None
    elif name == "maps_round1":
        # anchor
        return f"^{PRIMER_MAPS_1};required...{get_rc(PRIMER_MAPS_r)};optional", get_rc(
            PRIMER_MAPS_1
        )
    elif name == "maps_round2":
        # anchor
        return f"^{PRIMER_MAPS_2}", None
    elif name == "maps_3":
        # anchor
        return f"^{PRIMER_MAPS_3}", None
    elif name == "maps_rand_hex_test":
        primer_maps = PRIMER_MAPS_1 + "N" * 6
        return (
            f"^{primer_maps};required...{get_rc(PRIMER_MAPS_r)};optional",
            get_rc(primer_maps),
        )
    elif name == "recording_adapter":
        return (
            f"{RECORDING_PRIMER_5};rightmost;min_overlap={get_min_overlap(RECORDING_PRIMER_5, frac=0.5)}"
            f"...{RECORDING_PRIMER_3};min_overlap={get_min_overlap(RECORDING_PRIMER_3, frac=0.5)}"
        )
    elif name == "recording_leader_dr":
        ret = ECREC_LEADER + ECREC_DR
        return ret + f";min_overlap={get_min_overlap(ret)}"
    elif name == "recording_spacer0":
        return (
            ECREC_SPACER0 + f";min_overlap={get_min_overlap(ECREC_SPACER0)};rightmost"
        )
    elif name == "recording_spacer0_rc":
        return get_rc(ECREC_SPACER0) + f";min_overlap={get_min_overlap(ECREC_SPACER0)}"
    elif name == "recording_dr":
        return ECREC_DR + f";min_overlap={get_min_overlap(ECREC_DR)}"
    elif name == "recording_dr_rc":
        return get_rc(ECREC_DR) + f"X;min_overlap={get_min_overlap(ECREC_DR)}"
    else:
        raise ValueError(f"Invalid primer set name: {name}")


def isolate_150_preprocess(
    fastq_dir: str,
    barcode_fwd_fasta: str,
    barcode_rev_fasta: str,
    rename_pattern: str,
    output_fastq: str,
    primer_set: str,
    first_k: int | None = None,
    min_length: int = 100,
    early_stop: bool = False,
) -> None:
    output_dir, output_f = os.path.split(output_fastq)
    output_dir_demux = os.path.join(output_dir, "demux")
    output_dir_demux_fail = os.path.join(output_dir, "demux_failed")
    output_dir_cutadapt = os.path.join(output_dir, "cutadapt")
    output_fastq_r1 = "_R1.".join(output_f.split(".", 1))
    output_fastq_r2 = "_R2.".join(output_f.split(".", 1))
    os.makedirs(output_dir, exist_ok=True)
    os.makedirs(output_dir_demux, exist_ok=True)
    os.makedirs(output_dir_demux_fail, exist_ok=True)
    os.makedirs(output_dir_cutadapt, exist_ok=True)

    # demultiplex with cutadapt
    proc_args = [
        "cutadapt",
        "-e",
        "0.15",
        "--no-indels",
        "-g",
        f"^file:{barcode_fwd_fasta}",
        "-G",
        f"^file:{barcode_rev_fasta}",
        "-o",
        os.path.join(output_dir_demux, "{name1}-{name2}_R1.fq.gz"),
        "-p",
        os.path.join(output_dir_demux, "{name1}-{name2}_R2.fq.gz"),
        "--cores",
        "1",
        "--interleaved",
        "-",
    ]
    print_command(proc_args)
    cutadapt_demux_proc = subprocess.Popen(
        proc_args,
        stdin=subprocess.PIPE,
        # silence cutadapt's output
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    cat_fastq(
        fastq_dir,
        output_fp_r1=cutadapt_demux_proc.stdin,
        output_fp_r2=cutadapt_demux_proc.stdin,
    )
    cutadapt_demux_proc.stdin.close()
    cutadapt_demux_proc.wait()
    rename_files_with_mmv(output_dir_demux, rename_pattern)

    # move all remaining files (whose name1 or name2 is "unknown") to output_dir/demux_failed
    demux_failed_files = list(
        set(
            glob.glob(os.path.join(output_dir_demux, "*-unknown_R1.fq.gz"))
            + glob.glob(os.path.join(output_dir_demux, "*-unknown_R2.fq.gz"))
            + glob.glob(os.path.join(output_dir_demux, "unknown-*_R1.fq.gz"))
            + glob.glob(os.path.join(output_dir_demux, "unknown-*_R2.fq.gz"))
        )
    )
    subprocess.run(
        [
            "mv",
            *demux_failed_files,
            os.path.join(output_dir, "demux_failed"),
        ]
    )

    # trim with cutadapt
    a, A = get_primer_set(primer_set)
    proc_args = (
        ["cutadapt", "-e", "0.15", "-a", a]
        + (["-A", A] if not early_stop else [])  # no trimming for read 2 if early stop
        + [
            "--minimum-length",
            str(min_length),
            "--pair-filter",
            "any",
            "-O",
            "16",
            "-o",
            os.path.join(output_dir_cutadapt, output_fastq_r1),
            "-p",
            os.path.join(output_dir_cutadapt, output_fastq_r2),
            "--untrimmed-output",
            os.path.join(output_dir_cutadapt, "untrimmed_1.fq.gz"),
            "--untrimmed-paired-output",
            os.path.join(output_dir_cutadapt, "untrimmed_2.fq.gz"),
            "--too-short-output",
            os.path.join(output_dir_cutadapt, "too_short_1.fq.gz"),
            "--too-short-paired-output",
            os.path.join(output_dir_cutadapt, "too_short_2.fq.gz"),
            "--cores",
            "4",
            "--interleaved",
            "-",
        ]
    )
    if first_k is not None:
        proc_args.extend(["-l", str(first_k)])
    print_command(proc_args)
    cutadapt_trim_proc = subprocess.Popen(
        proc_args,
        stdin=subprocess.PIPE,
        # silence cutadapt's output
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    cat_fastq(
        output_dir_demux,
        output_fp_r1=cutadapt_trim_proc.stdin,
        output_fp_r2=cutadapt_trim_proc.stdin,
        _have_sample_name=True,
    )
    cutadapt_trim_proc.stdin.close()
    cutadapt_trim_proc.wait()

    # copy merged_1 to output_fastq
    shutil.copy(os.path.join(output_dir_cutadapt, output_fastq_r1), output_fastq)
    shutil.rmtree(output_dir_demux)


def cutadapt_demux_se(fastq_path: str, output_dir: str, barcode_fastq: str):
    if not os.path.isfile(barcode_fastq):
        raise ValueError(f"{barcode_fastq} does not exist.")
    output_dir_cutadapt = os.path.join(output_dir, "cutadapt")
    output_dir_demux = os.path.join(output_dir, "demux")
    output_dir_demux_fail = os.path.join(output_dir, "demux_failed")
    os.makedirs(output_dir, exist_ok=True)
    os.makedirs(output_dir_demux, exist_ok=True)
    os.makedirs(output_dir_demux_fail, exist_ok=True)
    os.makedirs(output_dir_cutadapt, exist_ok=True)
    # demultiplex with cutadapt
    proc_args = [
        "cutadapt",
        "-e",
        "0.2",
        "--no-indels",
        "-g",
        f"^file:{barcode_fastq}",
        "-o",
        os.path.join(output_dir_demux, "{name}.fq.gz"),
    ]
    proc_args.append(fastq_path)
    subprocess.run(proc_args)
    subprocess.run(
        [
            "mv",
            os.path.join(output_dir_demux, "unknown.fq.gz"),
            os.path.join(output_dir, "demux_failed"),
        ]
    )


def cutadapt_demux_merge_trim_se(
    fastq_path: str,
    output_fastq: str,
    primer_set: str,
    barcode_fastq: str,
    first_k: int | None = None,
    min_length: int | None = None,
    _r2: bool = False,
) -> None:
    """Demultiplex and merge single-end reads using cutadapt."""
    if not os.path.isfile(barcode_fastq):
        raise ValueError(f"{barcode_fastq} does not exist.")
    output_dir, output_f = os.path.split(output_fastq)
    output_dir_cutadapt = os.path.join(output_dir, "cutadapt")
    output_dir_demux = os.path.join(output_dir, "demux")
    output_dir_demux_fail = os.path.join(output_dir, "demux_failed")
    os.makedirs(output_dir, exist_ok=True)
    os.makedirs(output_dir_demux, exist_ok=True)
    os.makedirs(output_dir_demux_fail, exist_ok=True)
    os.makedirs(output_dir_cutadapt, exist_ok=True)

    # demultiplex with cutadapt
    proc_args = [
        "cutadapt",
        "-e",
        "1",
        "--no-indels",
        "-g",
        f"^file:{barcode_fastq}",
        "-o",
        os.path.join(output_dir_demux, "{name}.fq.gz"),
    ]
    if os.path.isdir(fastq_path):
        proc_args.append("-")
        print_command(proc_args)
        cutadapt_demux_proc = subprocess.Popen(
            proc_args,
            stdin=subprocess.PIPE,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        cat_fastq_se(
            fastq_path,
            output_fp=cutadapt_demux_proc.stdin,
        )
        cutadapt_demux_proc.stdin.close()
        cutadapt_demux_proc.wait()
    else:
        proc_args.append(fastq_path)
        print_command(proc_args)
        subprocess.run(proc_args)

    # no need for renaming
    # rename_files_with_mmv(output_dir_demux, rename_pattern)

    subprocess.run(
        [
            "mv",
            os.path.join(output_dir_demux, "unknown.fq.gz"),
            os.path.join(output_dir, "demux_failed"),
        ]
    )

    a, _ = get_primer_set(primer_set)
    cutadapt_trim_proc_args = [
        "cutadapt",
        # "-a" if "..." in a else "-g",
        "-g",
        a,
        "-o",
        output_fastq,
        "--untrimmed-output",
        os.path.join(output_dir_cutadapt, "untrimmed.fq.gz"),
        # "--too-short-output",
        # os.path.join(output_dir_cutadapt, "too_short.fq.gz"),
        "--cores",
        "8",
        "-",
    ]
    if first_k is not None:
        cutadapt_trim_proc_args.extend(["-l", str(first_k)])
    if min_length is not None:
        cutadapt_trim_proc_args.extend(
            [
                "--minimum-length",
                str(min_length),
                "--too-short-output",
                os.path.join(output_dir_cutadapt, "too_short_1.fq.gz"),
            ]
        )
    print_command(cutadapt_trim_proc_args)
    cutadapt_trim_proc = subprocess.Popen(
        cutadapt_trim_proc_args,
        stdin=subprocess.PIPE,
        # silence cutadapt's output
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    cat_fastq_se(
        output_dir_demux,
        output_fp=cutadapt_trim_proc.stdin,
        _have_sample_name=True,
        _r2=_r2,
    )
    cutadapt_trim_proc.stdin.close()
    cutadapt_trim_proc.wait()

    shutil.rmtree(output_dir_demux)


def cutadapt_demux_merge_trim_pe(
    fastq_path: str, output_dir: str, primer_set: str, barcode_fastq: str
) -> None:
    """Demultiplex and merge paired-end reads using cutadapt."""
    if not os.path.isfile(barcode_fastq):
        raise ValueError(f"{barcode_fastq} does not exist.")
    output_dir_cutadapt = os.path.join(output_dir, "cutadapt")
    output_dir_demux = os.path.join(output_dir, "demux")
    output_dir_demux_fail = os.path.join(output_dir, "demux_failed")
    os.makedirs(output_dir, exist_ok=True)
    os.makedirs(output_dir_demux, exist_ok=True)
    os.makedirs(output_dir_demux_fail, exist_ok=True)
    os.makedirs(output_dir_cutadapt, exist_ok=True)

    # demultiplex with cutadapt
    proc_args = [
        "cutadapt",
        "-e",
        "1",
        "--no-indels",
        "-g",
        f"^file:{barcode_fastq}",
        "-o",
        os.path.join(output_dir_demux, "{name}_R1.fq.gz"),
        "-p",
        os.path.join(output_dir_demux, "{name}_R2.fq.gz"),
    ]
    if os.path.isdir(fastq_path):
        proc_args.extend(["--interleaved", "-"])
        print_command(proc_args)
        cutadapt_demux_proc = subprocess.Popen(
            proc_args,
            stdin=subprocess.PIPE,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        cat_fastq(
            fastq_path,
            output_fp_r1=cutadapt_demux_proc.stdin,
            output_fp_r2=cutadapt_demux_proc.stdin,
        )
        cutadapt_demux_proc.stdin.close()
        cutadapt_demux_proc.wait()
    else:
        print_command(proc_args)
        proc_args.extend([fastq_path])
        subprocess.run(proc_args)

    # no need for renaming
    # rename_files_with_mmv(output_dir_demux, rename_pattern)
    subprocess.run(
        [
            "mv",
            os.path.join(output_dir_demux, "unknown_R1.fq.gz"),
            os.path.join(output_dir_demux, "unknown_R2.fq.gz"),
            os.path.join(output_dir, "demux_failed"),
        ]
    )
    a, A = get_primer_set(primer_set)
    proc_args = [
        "cutadapt",
        "-e",
        "0.15",
        "-a",
        a,
        "-A",
        A,
        "--pair-filter",
        "first",  # discard pair if first read is not trimmed
        "-o",
        os.path.join(output_dir, "merged_R1.fq.gz"),
        "-p",
        os.path.join(output_dir, "merged_R2.fq.gz"),
        "--untrimmed-output",
        os.path.join(output_dir_cutadapt, "untrimmed_1.fq.gz"),
        "--untrimmed-paired-output",
        os.path.join(output_dir_cutadapt, "untrimmed_2.fq.gz"),
        "--cores",
        "4",
        "--interleaved",
        "-",
    ]
    print_command(proc_args)
    cutadapt_trim_proc = subprocess.Popen(
        proc_args,
        stdin=subprocess.PIPE,
        # silence cutadapt's output
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    cat_fastq(
        output_dir_demux,
        output_fp_r1=cutadapt_trim_proc.stdin,
        output_fp_r2=cutadapt_trim_proc.stdin,
        _have_sample_name=True,
    )
    cutadapt_trim_proc.stdin.close()
    cutadapt_trim_proc.wait()

    shutil.rmtree(output_dir_demux)


def cutadapt_merge_trim_se(
    fastq_path: str,
    output_fastq: str,
    primer_set: str,
    first_k: int | None = None,
    min_length: int | None = None,
    _r2: bool = False,
) -> None:
    output_dir, output_f = os.path.split(output_fastq)
    os.makedirs(output_dir, exist_ok=True)
    a, A = get_primer_set(primer_set)
    proc_args = [
        "cutadapt",
        "-g" if primer_set.endswith("_5") else "-a",
        a,
        "-o",
        output_fastq,
        "--cores",
        "8",
    ]
    if first_k is not None:
        proc_args.extend(["-l", str(first_k)])
    if min_length is not None:
        output_dir_cutadapt = os.path.join(output_dir, "cutadapt")
        os.makedirs(output_dir_cutadapt, exist_ok=True)
        proc_args.extend(
            [
                "--minimum-length",
                str(min_length),
                "--too-short-output",
                os.path.join(output_dir_cutadapt, "too_short_1.fq.gz"),
            ]
        )
    if os.path.isdir(fastq_path):
        proc_args.append("-")
        cutadapt_trim_proc = subprocess.Popen(
            proc_args,
            stdin=subprocess.PIPE,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        cat_fastq_se(fastq_path, output_fp=cutadapt_trim_proc.stdin, _r2=_r2)
        cutadapt_trim_proc.stdin.close()
        cutadapt_trim_proc.wait()
    else:
        proc_args.append(fastq_path)
        subprocess.run(proc_args)


def extract_r2_16s_v3v4(fastq_dir: str, output_fastq: str) -> None:
    length = 150
    output_dir, output_f = os.path.split(output_fastq)
    os.makedirs(output_dir, exist_ok=True)
    log_dir = os.path.join(output_dir, "fastp")
    args = (
        f"fastp -w 8 "
        f"--length_required {length} "
        f"--cut_right "
        f"--stdin "
        f"--stdout "
        f"--json {log_dir}/report.json "
        f"--html {log_dir}/report.html "
        f"| seqkit seq -rp "
        f"| cutadapt -o {output_fastq} -l {length} -"
    )
    print_command(args)
    preprocess_proc = subprocess.Popen(
        args,
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    cat_fastq_se(fastq_dir, output_fp=preprocess_proc.stdin, _r2=True)
    preprocess_proc.stdin.close()
    preprocess_proc.wait()


def merge(fastq_dir: str, output_dir: str) -> None:
    """Take output directory of Illumina BCL Convert, merge the fastq files and rename
    each read to `@<sample_name> <read_index> <read 1 or read 2>`. Output are written to
    `output_dir/merged_R1.fastq.gz` and `output_dir/merged_R2.fastq.gz`.

    Note:
        The name of the fastq files are supposed to follow the naming convention of
            `<sample_name>_S<sample_number>_L001_R[1/2]_001.fastq.gz`

    Concatenate files by using subprocess to run something like this:

    for fq1 in "$RAW_DATA_DIR"/*_1.fq; do
        fq2="${fq1%_1.fq}_2.fq"
        sample_name=$(basename "$fq1" "_1.fq")

        awk -v sn="$sample_name" 'BEGIN {
            read_index=1
        }
        {
            if(NR%4==1) {
                print "@sample=" sn " " read_index " 1"
                read_index++
            } else {
                print $0
            }
        }' "$fq1" | gzip >> "$TEMP_DIR/all_samples_1.fq.gz"

        awk -v sn="$sample_name" 'BEGIN {
            read_index=1
        }
        {
            if(NR%4==1) {
                print "@sample=" sn " " read_index " 2"
                read_index++
            } else {
                print $0
            }
        }' "$fq2" | gzip >> "$TEMP_DIR/all_samples_2.fq.gz"

    """
    r1_files = glob.glob(os.path.join(fastq_dir, "*_L001_R1_001.fastq.gz"))
    r2_files = glob.glob(os.path.join(fastq_dir, "*_L001_R2_001.fastq.gz"))
    r1_sample_names = [os.path.basename(f).split("_")[0] for f in r1_files]
    r2_sample_names = [os.path.basename(f).split("_")[0] for f in r2_files]
    # check pairing and uniqueness
    if r1_sample_names != r2_sample_names:
        raise ValueError("Sample names in R1 and R2 do not match.")
    if len(set(r1_sample_names)) != len(r1_sample_names):
        raise ValueError("Sample names are not unique.")

    out_r1 = os.path.join(output_dir, "merged_R1.fastq.gz")
    out_r2 = os.path.join(output_dir, "merged_R2.fastq.gz")
    # remove old files if exist
    if os.path.isfile(out_r1):
        os.remove(out_r1)
    if os.path.isfile(out_r2):
        os.remove(out_r2)
    os.makedirs(output_dir, exist_ok=True)

    # merge
    for r1_file, r2_file, sample_name in zip(r1_files, r2_files, r1_sample_names):
        # merge R1
        for read_num, fq_in, fq_out in zip(
            [1, 2], [r1_file, r2_file], [out_r1, out_r2]
        ):
            process = [
                "awk",
                "-v",
                f"sn={sample_name}",
                f"""'BEGIN {{ read_index=1 }} {{ if(NR%4==1) {{ print "@sample=" sn " " read_index " {read_num}"; read_index++ }} else {{ print $0 }} }}'""",
                fq_in,
                "|",
                "gzip",
                ">>",
                fq_out,
            ]
            subprocess.run(" ".join(process), shell=True)


def _merge_fastq_gz(r1: str, r2: str, output: str) -> None:
    """
    Merges two paired-end gzipped FASTQ files by concatenating the sequence and quality lines.

    Args:
        r1 (str): Path to the first gzipped FASTQ file (R1).
        r2 (str): Path to the second gzipped FASTQ file (R2).
        output (str): Path to the output merged gzipped FASTQ file.
    """
    if not r1.endswith(".gz") or not r2.endswith(".gz") or not output.endswith(".gz"):
        raise ValueError("Input and output files must be gzipped FASTQ files.")
    cmd = f"""
paste -d'\\n' \
<(zcat {r1} | awk 'NR%4==1') \
<(paste -d '' <(zcat {r1} | awk 'NR%4==2') <(zcat {r2} | awk 'NR%4==2')) \
<(zcat {r1} | awk 'NR%4==3') \
<(paste -d '' <(zcat {r1} | awk 'NR%4==0') <(zcat {r2} | awk 'NR%4==0')) \
| gzip -c > {output}"""
    print_command(cmd)

    subprocess.run(cmd, shell=True, executable="/bin/bash", check=True)


#     cmd = f"""
# paste <(zcat {r1} | paste - - - -) <(zcat {r2} | paste - - - -) |
# awk -v OFS="\\n" '{{print $1, $2$6, $3, $4$8}}' | gzip > {output}"""

#     try:
#         subprocess.run(cmd, shell=True, check=True, executable="/bin/bash")
#     except subprocess.CalledProcessError as e:
#         raise RuntimeError(f"Error merging FASTQ files: {e}")


def rename_output_files(output_dir, pair):
    """Rename the output files generated by trim_galore."""
    r1_old = os.path.join(output_dir, pair[0].replace("_R1.fastq", "_R1_val_1.fq"))
    r1_new = os.path.join(output_dir, pair[0])
    if not os.path.isfile(r1_old):
        print(
            f"ERROR: {r1_old} does not exist. This is not expected and should be investigated."
        )
    else:
        os.rename(r1_old, r1_new)

    r2_old = os.path.join(output_dir, pair[1].replace("_R2.fastq", "_R2_val_2.fq"))
    r2_new = os.path.join(output_dir, pair[1])
    if not os.path.isfile(r2_old):
        print(
            f"ERROR: {r2_old} does not exist. This is not expected and should be investigated."
        )
    else:
        os.rename(r2_old, r2_new)


def resolve_input_path(input_path: str) -> str:
    """Resolve the input path relative to the 'data' directory if not an existing file.

    Args:
        input_path: A path string or None. If not None and not a file, assume it is relative to the 'data' directory.

    Returns:
        The resolved absolute path or None.
    """
    if not os.path.isfile(input_path):
        return os.path.join(os.path.dirname(__file__), "data", input_path)
    return input_path


# if __name__ == "__main__":
def main():
    parser = argparse.ArgumentParser(
        description="Process and rename FASTQ files using trim_galore."
    )
    parser.add_argument(
        "-i",
        "--input_dir",
        type=str,
        help="Input directory containing FASTQ files",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Output FASTQ file (could be gzipped) path or directory",
        required=True,
    )
    parser.add_argument(
        "-p",
        "--primer_set",
        type=str,
        choices=["its", "16s", "maps_round1", "maps_round2"],
        help="Primer set used",
    )
    parser.add_argument(
        "-fb",
        "--barcode_fwd",
        type=str,
        default=None,
        help="Barcode fasta file for forward reads",
    )
    parser.add_argument(
        "-rb",
        "--barcode_rev",
        type=str,
        default=None,
        help="Barcode fasta file for reverse reads",
    )
    parser.add_argument(
        "-pt",
        "--pattern",
        type=str,
        help="Pattern file for mmv, used to rename the files",
    )
    parser.add_argument(
        "-m",
        "--mode",
        type=str,
        default="simple",
        choices=[
            "simple",
            "pseudo_merge",
            "r1",
            "r1_demux_trim",
            "isolate_150",
            "isolate_150_early_stop",
            "maps_round0",
            "maps_round1",
            "maps_round2",
            "maps_round3",
            "maps_rand_hex_test",
        ],
        help="Processing mode",
    )
    parser.add_argument(
        "-k", "--first_k", type=int, default=None, help="The first k bases to keep"
    )
    parser.add_argument(
        "-l", "--min_length", type=int, default=None, help="Minimum length to keep"
    )
    parser.add_argument(
        "-k2", "--first_k2", type=int, default=None, help="The first k bases to keep"
    )
    parser.add_argument(
        "-l2", "--min_length2", type=int, default=None, help="Minimum length to keep"
    )

    args = parser.parse_args()

    if args.mode == "simple":
        simple_preprocess(args.input_dir, args.output)
    elif args.mode == "pseudo_merge":
        # take first_k1 for read1, and first_k2 for read2, then merge them.
        pseudo_merge(
            args.input_dir,
            args.output,
            threads=8,
            min_l1=args.min_length,
            first_k1=args.first_k,
            min_l2=args.min_length2,
            first_k2=args.first_k2,
        )
    elif args.mode == "isolate_150":
        # for barcode_fwd, barcode_rev and pattern, first look for them as if they are
        # normal file path, if not exist, then look for them in script_dir/../../data.
        barcode_fwd = resolve_input_path(args.barcode_fwd)
        barcode_rev = resolve_input_path(args.barcode_rev)
        pattern = resolve_input_path(args.pattern)
        isolate_150_preprocess(
            args.input_dir,
            barcode_fwd,
            barcode_rev,
            pattern,
            args.output,
            primer_set=args.primer_set,
            first_k=args.first_k,
            min_length=args.min_length,
        )
    elif args.mode == "isolate_150_early_stop":
        isolate_150_preprocess(
            args.input_dir,
            args.barcode_fwd,
            args.barcode_rev,
            args.pattern,
            args.output,
            primer_set=args.primer_set,
            first_k=args.first_k,
            min_length=args.min_length,
            early_stop=True,
        )
    elif args.mode == "r1":
        cutadapt_merge_trim_se(
            args.input_dir,
            args.output,
            primer_set=args.primer_set,
            first_k=args.first_k,
            min_length=args.min_length,
        )
    elif args.mode == "r1_demux_trim":
        barcode_fwd = resolve_input_path(args.barcode_fwd)
        cutadapt_demux_merge_trim_se(
            args.input_dir,
            args.output,
            primer_set=args.primer_set,
            barcode_fastq=barcode_fwd,
            first_k=args.first_k,
            min_length=args.min_length,
        )
    elif args.mode == "maps_round1":
        cutadapt_demux_merge_trim_se(
            args.input_dir,
            args.output,
            barcode_fastq=args.barcode_fwd,
            primer_set="maps_1",
        )
    elif args.mode == "maps_round2":
        cutadapt_demux_merge_trim_se(
            args.input_dir,
            args.output,
            barcode_fastq=args.barcode_fwd,
            primer_set="maps_2",
        )
    elif args.mode == "maps_round3":
        cutadapt_demux_merge_trim_se(
            args.input_dir,
            args.output,
            barcode_fastq=args.barcode_fwd,
            primer_set="maps_3",
        )
    elif args.mode == "maps_rand_hex_test":
        cutadapt_demux_merge_trim_pe(
            args.input_dir,
            args.output,
            barcode_fastq=args.barcode_fwd,
            primer_set="maps_rand_hex_test",
        )



import io
import glob
import gzip
import os
import re
import warnings
from typing import IO

import pandas as pd
import biom
from tqdm.auto import tqdm


PATTERN_ILLUMINA = re.compile(r"^(.+?)_S\d+_(?:L\d{3}_)?(R[12])_001.f(ast)?q(.gz)?$")
# - sample_R1.fq.gz or sample_R2.fq.gz (or fastq.gz, same for the following)
# - sample_1.fq.gz or sample_2.fq.gz
# - sample.R1.fq.gz or sample.R2.fq.gz
# - sample.1.fq.gz or sample.2.fq.gz
PATTERN_CUSTOM = re.compile(r"^(.+?)(?:(?:_|\.)(?:R)?[12])?.f(?:ast)?q(?:.gz)?$")


def find_paired_end_files(directory: str) -> list[tuple[str, str, str]]:
    # Dictionary to store file pairs
    file_pairs = {}

    # Find all .fastq.gz and .fq.gz files in the directory
    fastq_files = (
        glob.glob(os.path.join(directory, "*.fastq.gz"))
        + glob.glob(os.path.join(directory, "*.fq.gz"))
        + glob.glob(os.path.join(directory, "*.fastq"))
        + glob.glob(os.path.join(directory, "*.fq"))
    )

    # Process each file using pattern functions
    for file_ in fastq_files:
        for pattern in [PATTERN_ILLUMINA, PATTERN_CUSTOM]:
            match = pattern.search(os.path.basename(file_))
            if match:
                sample_name, read_type = match.groups()[:2]
                read_type = {"1": "R1", "2": "R2"}.get(read_type, read_type)
                if read_type not in ["R1", "R2"]:
                    warnings.warn(f"Invalid read type for file: {file_}")
                    continue

                pair_key = (sample_name, read_type)

                # Add the file to the dictionary and check for duplicates
                if pair_key in file_pairs:
                    warnings.warn(f"Duplicate file for {pair_key}: {file_}")
                else:
                    file_pairs[pair_key] = file_
                break
        else:
            warnings.warn(f"File {file_} does not match any known pattern.")

    # Match R1 and R2 pairs and prepare the output
    matched_pairs = []
    processed_samples = set()

    for (sample_name, read_type), file_ in file_pairs.items():
        if sample_name in processed_samples:
            # Skip if we've already processed this sample
            continue

        processed_samples.add(sample_name)
        pair_type = "R2" if read_type == "R1" else "R1"
        pair_key = (sample_name, pair_type)

        if pair_key in file_pairs:
            r1_file = file_ if read_type == "R1" else file_pairs[pair_key]
            r2_file = file_pairs[pair_key] if read_type == "R1" else file_
            matched_pairs.append((r1_file, r2_file, sample_name))
        else:
            warnings.warn(f"Missing pair for file: {file_}")

    # Report the number of matched pairs
    print(f"Number of matched pairs: {len(matched_pairs)}")

    # Order by sample name
    matched_pairs.sort(key=lambda x: x[2])
    return matched_pairs


def smart_open(file_path: str, mode: str = "r") -> IO[str] | gzip.GzipFile:
    """Open a file as text or as a gzip file based on its magic number.

    Args:
        file_path (str): Path to the file to be opened.
        mode (str): Mode in which the file should be opened. Defaults to 'rt' (read text).

    Returns:
        IO[str] | gzip.GzipFile: A file object or a gzip file object.
    """

    if mode == "r":
        with open(file_path, "rb") as f:
            first_two_bytes = f.read(2)
        if first_two_bytes == b"\x1f\x8b":  # Magic number for gzip files
            return gzip.open(file_path, "rt")
        else:
            return open(file_path, "r")
    elif mode == "w":
        if file_path.endswith(".gz") or file_path.endswith(".gzip"):
            return gzip.open(file_path, "wt")
        else:
            return open(file_path, "w")
    else:
        raise ValueError(f"Mode must be 'r' or 'w' if specified, getting {mode}.")


def cat_fastq(
    directory: str,
    output_fp_r1,
    output_fp_r2=None,
    metadata: str | None = None,
    _remove_undet: bool = True,
    _have_sample_name: bool = False,
) -> None:
    """Process FASTQ files in the given directory, renaming reads,and write the output
    to the specified file pointers. Output fastq will be interleaved if `output_fp_r2`
    is None.

    This function reinvents the wheel implemented in many bioinformatics tools, but I
    am still doing this for customizing the read renaming.

    Args:
        directory: Directory containing FASTQ files.
        output_fp_r1: File pointer to write the R1 output.
        output_fp_r2: File pointer to write the R2 output.
    """
    matched_pairs = find_paired_end_files(directory)
    if output_fp_r2 is None:
        output_fp_r2 = output_fp_r1
    if metadata is not None:
        samples_in_meta = pd.read_table(metadata, index_col=0).index.to_list()
    else:
        samples_in_meta = None

    if (
        (
            isinstance(output_fp_r1, gzip.GzipFile)
            and isinstance(output_fp_r2, gzip.GzipFile)
        )
        or (
            isinstance(output_fp_r1, io.BufferedWriter)
            and isinstance(output_fp_r2, io.BufferedWriter)
        )
        or (
            isinstance(output_fp_r1, io.BytesIO)
            and isinstance(output_fp_r2, io.BytesIO)
        )
    ):

        def write_line(line1, line2: str):
            output_fp_r1.write(line1.encode())
            output_fp_r2.write(line2.encode())

    elif isinstance(output_fp_r1, io.TextIOBase) and isinstance(
        output_fp_r2, io.TextIOBase
    ):

        def write_line(line1, line2: str):
            output_fp_r1.write(line1)
            output_fp_r2.write(line2)

    else:
        raise ValueError("Output file pointers must be both gzip or both text file.")

    if _have_sample_name:
        rename_read = _rename_read_concat
    else:
        rename_read = _rename_read_illumina

    for idx, (r1_path, r2_path, sample_name) in enumerate(tqdm(matched_pairs)):
        if samples_in_meta is not None and sample_name not in samples_in_meta:
            continue
        if _remove_undet and sample_name == "Undetermined":
            continue
        # if idx < 4:
        #     continue
        # print(sample_name)
        with smart_open(r1_path) as r1_file, smart_open(r2_path) as r2_file:
            paired_read_iter = zip(
                zip(*[r1_file] * 4, strict=True),
                zip(*[r2_file] * 4, strict=True),
                strict=True,
            )

            for read_index, (r1_lines, r2_lines) in enumerate(
                paired_read_iter, start=1
            ):
                # Renaming reads
                write_line(
                    rename_read(r1_lines[0], sample_name, 1, read_index)
                    + "".join(r1_lines[1:]),
                    rename_read(r2_lines[0], sample_name, 2, read_index)
                    + "".join(r2_lines[1:]),
                )


def cat_fastq_se(
    directory: str,
    output_fp,
    metadata: str | None = None,
    _remove_undet: bool = True,
    _have_sample_name: bool = False,
    _r2: bool = False,
):
    """Similar as above but simply list all fastq/fq/fastq.gz/fq.gz files in the
    directory and concatenate them into a single file with new read names. Good for
    single-end reads.
    """
    fastq_files = (
        glob.glob(os.path.join(directory, "*.fastq.gz"))
        + glob.glob(os.path.join(directory, "*.fq.gz"))
        + glob.glob(os.path.join(directory, "*.fastq"))
        + glob.glob(os.path.join(directory, "*.fq"))
    )

    files = []
    for file_ in fastq_files:
        for pattern in [PATTERN_ILLUMINA, PATTERN_CUSTOM]:
            match = pattern.search(os.path.basename(file_))
            if match:
                sample_name, read_type = match.groups()[:2]
                read_type = {None: "R1", "1": "R1", "2": "R2"}.get(read_type, read_type)
                if read_type not in ["R1", "R2"]:
                    warnings.warn(f"Invalid read type {read_type} for file: {file_}")
                    continue
                if _r2:
                    if read_type == "R1":
                        break
                else:
                    if read_type == "R2":
                        break
                files.append((file_, sample_name, read_type))
                break

    if metadata is not None:
        samples_in_meta = pd.read_table(metadata, index_col=0).index.to_list()
    else:
        samples_in_meta = None

    if isinstance(output_fp, gzip.GzipFile) or isinstance(output_fp, io.BufferedWriter):

        def write_line(line):
            output_fp.write(line.encode())

    elif isinstance(output_fp, io.TextIOBase):

        def write_line(line):
            output_fp.write(line)

    else:
        raise ValueError("Output file pointer must be gzip or text file.")

    if _have_sample_name:
        rename_read = _rename_read_concat
    else:
        rename_read = _rename_read_illumina

    prog_bar = tqdm(files)
    for file_, sample_name, read_type in prog_bar:
        # match = PATTERN_ILLUMINA.search(os.path.basename(file_))
        # if match:
        # sample_name = match.group(1)
        # sample_name = os.path.basename(file_).split(".")[0].split("_")[0]
        if samples_in_meta is not None and sample_name not in samples_in_meta:
            continue
        if _remove_undet and sample_name == "Undetermined":
            continue
        with smart_open(file_) as f:
            for read_index, lines in enumerate(zip(*[f] * 4, strict=True), start=1):
                write_line(
                    rename_read(lines[0], sample_name, 1, read_index)
                    + "".join(lines[1:])
                )


def _rename_read_illumina(
    header_line: str, sample_name: str, read_number: int, read_index: int
) -> str:
    """
    Rename a read header line with the sample name, read number, and read index.

    Args:
        header_line: Original header line from FASTQ file.
        sample_name: Name of the sample.
        read_number: Read number (1 or 2).
        read_index: Index of the read in the sample.

    Returns:
        Renamed header line.
    """
    # Remove '@' and split by space, take first part
    original_header = header_line.strip().split()[0][1:]
    return f"@sample={sample_name} {read_number} {read_index} {original_header}\n"


def _rename_read_concat(
    header_line: str, sample_name: str, read_number: int, read_index: int
) -> str:
    header_line = header_line[1:]
    original_header, comment = header_line.split(maxsplit=1)
    original_sample = original_header.split("=", 1)[1]
    new_sample = f"{original_sample}_{sample_name}"
    return f"@sample={new_sample} {comment}"


def print_command(command: str | list[str]) -> None:
    """Print the given command in a readable format.
    Iterate through the command list and print each element on a new line if starts with
    a dash, otherwise print it on the same line.
    """
    print("Running the command:")
    if isinstance(command, str):
        print(command)
        return
    else:
        print(command[0], end="")
        for arg in command[1:]:
            if arg.startswith("-"):
                print(" \\\n    " + arg, end="")
            else:
                print(" " + arg, end="")
        print()


def read_table(
    table_path: str,
    index_col: str | int = 0,
    comment: str = None,
    dtype: str = "int",
    index_name: str = None,
) -> pd.DataFrame:
    """Read a table from a file and return it as a DataFrame.

    Args:
        table_path: Path to the table file. Could be a tsv, csv or biom.

    Returns:
        pd.DataFrame: DataFrame containing the table data.
    """
    if table_path.endswith(".biom"):
        df_biom = biom.load_table(table_path)
        df = df_biom.to_dataframe().astype(dtype)
        df.index.name = df_biom.table_id if index_name is None else index_name
        return df
    else:
        if table_path.endswith(".csv") or table_path.endswith(".csv.gz"):
            method = pd.read_csv
        elif table_path.endswith(".tsv") or table_path.endswith(".tsv.gz"):
            method = pd.read_table
        else:
            raise ValueError("Unsupported table file format.")
        return method(
            table_path, index_col=index_col, comment=comment, dtype={index_col: str}
        )


def write_table(table: pd.DataFrame, table_path: str) -> None:
    """Write a table to a file.

    Args:
        table: DataFrame to be written to a file.
        table_path: Path to the output file. Could be a tsv, csv or biom.
    """
    if table_path.endswith(".biom"):
        data = biom.Table(
            table.to_numpy(), table.index, table.columns, table_id=table.index.name
        )
        with biom.util.biom_open(table_path, "w") as f:
            data.to_hdf5(f, "whatever60")
    elif table_path.endswith(".csv") or table_path.endswith(".csv.gz"):
        table.to_csv(table_path)
    elif table_path.endswith(".tsv") or table_path.endswith(".tsv.gz"):
        table.to_csv(table_path, sep="\t")
    else:
        raise ValueError("Unsupported table file format.")



"""
Take a OTU count table, sample metadata, and OTU taxonomy table as input, output 
tables of relative/absolute abundance at OTU level and relative abundance at 
genus/family level.

QC is also performed during the process. Poor samples with low total OTU counts are 
removed, and poor genus and family with low total relative abundance are removed.

Spike-in genus needs to be specified as an argument. Sample weight could also be 
provided as an argument. Spike-in genus won't be considered in the final tables. 

Absolute OTU abundance is calculated by dividing OTU counts by spike-in counts and 
sample weight.

Relative OTU abundance is calculated by normalizing absolute OTU abundance to 1.

Relative genus/family abundance is calculated by first averaging over any replicates 
within group and averaging again relative OTU abundance of OTUs corresponding to the 
same genus/family.

Note that rows of OTU abundance tables are sample, while rows of genus/family abundance 
tables are groups.

Example run:
python3 scripts/analysis.py \
  --count_tsv_path outputs/zotu_otus.tsv \
  --sample_metadata_path metadatas/all.tsv \
  --otu_taxonomy_path outputs_rrndb/rrndb_zotus_relabel_processed.tsv \
  --otus_spikein Otu8 Otu96 Otu235 \
  --save_dir outputs_rrndb/all/thres_1 \
  --group_key Group \
  --sample_count_thres 100 \
  --tax_level genus \
  --rel_ab_thres 0.01 \
  --keep_others \
  --keep_unknown
#   --genus_rel_ab_thres 0.05 \
#   --family_rel_ab_thres 0.05
#   --genus_key genus \
#   --family_key family \
"""

import os
import argparse

import pandas as pd
import polars as pl
from skbio.diversity.alpha import shannon


def main():
    parser = argparse.ArgumentParser()

    # Arguments
    parser.add_argument("--count_tsv_path", required=True, help="Path to OTU count tsv.")
    parser.add_argument(
        "--sample_metadata_path", required=True, help="Path to sample metadata."
    )
    parser.add_argument(
        "--otu_taxonomy_path", required=True, help="Path to OTU taxonomy tsv."
    )
    parser.add_argument("--otus_spikein", nargs="+", help="OTUs spikein.", default=[])
    parser.add_argument("--sample_weight_key", default=None)
    parser.add_argument("--save_dir", required=True, help="Directory to save the outputs.")
    parser.add_argument("--group_key", default=None)
    parser.add_argument(
        "--sample_count_thres", type=int, required=True, help="Sample count threshold."
    )
    parser.add_argument(
        "--tax_level",
        type=str,
        required=True,
        help="Taxonomy level of interest, must be a column in the taxonomy table.",
    )
    parser.add_argument(
        "--rel_ab_thres",
        type=float,
        required=True,
        help="Genus relative abundance threshold.",
    )
    # parser.add_argument("--genus_key", required=True, help="Key for genus.")
    # parser.add_argument("--family_key", required=True, help="Key for family.")
    # parser.add_argument(
    #     "--genus_rel_ab_thres",
    #     type=float,
    #     required=True,
    #     help="Genus relative abundance threshold.",
    # )
    # parser.add_argument(
    #     "--family_rel_ab_thres",
    #     type=float,
    #     required=True,
    #     help="Family relative abundance threshold.",
    # )
    parser.add_argument("--keep_others", action="store_true", default=False)
    parser.add_argument("--keep_unknown", action="store_true", default=False)
    args = parser.parse_args()

    # Extract arg values
    count_tsv_path = args.count_tsv_path
    sample_metadata_path = args.sample_metadata_path
    otu_taxonomy_path = args.otu_taxonomy_path
    otus_spikein = args.otus_spikein
    sample_weight_key = args.sample_weight_key
    save_dir = args.save_dir
    # genus_key = args.genus_key
    # family_key = args.family_key
    group_key = args.group_key
    sample_count_thres = args.sample_count_thres
    tax_level = args.tax_level
    rel_ab_thres = args.rel_ab_thres
    # genus_rel_ab_thres = args.genus_rel_ab_thres
    # family_rel_ab_thres = args.family_rel_ab_thres
    keep_others = args.keep_others
    keep_unknown = args.keep_unknown

    os.makedirs(save_dir, exist_ok=True)

    # make sure the first column is sample id
    sample_metadata = pd.read_table(sample_metadata_path, index_col=0)
    if sample_weight_key is None:
        sample_weight_key = "sample_weight"
        sample_metadata[sample_weight_key] = 1
    if group_key is None:
        sample_metadata[group_key] = sample_metadata.index


    otu_taxonomy = pd.read_table(otu_taxonomy_path, index_col="otu")
    # reorder data_count columns by sample metadata
    # num_samples x num_otus
    # exclude OTUs without taxonomy assignment.
    data_count = pd.read_table(count_tsv_path, index_col=0)[sample_metadata.index].T
    common_otus = list(set(data_count.columns).intersection(otu_taxonomy.index))
    data_count = data_count[common_otus]
    otu_taxonomy = otu_taxonomy.loc[common_otus]
    pass_qc = data_count.sum(axis=1) >= sample_count_thres
    sample_metadata["total_reads"] = data_count.drop(otus_spikein, axis=1).sum(axis=1)
    sample_metadata["spikein_reads"] = data_count.loc[:, otus_spikein].sum(axis=1)

    # filter poor samples
    sample_metadata = sample_metadata.loc[pass_qc]
    data_count = data_count.loc[pass_qc].drop(otus_spikein, axis=1)

    if not otus_spikein:
        data_abs = data_count.div(sample_metadata[sample_weight_key], axis=0)
    else:
        data_abs = data_count.div(
            sample_metadata.spikein_reads * sample_metadata[sample_weight_key], axis=0
        )

    # for relative abundance, spike-in and sample weight do not matter.
    data_rel = data_count.div(sample_metadata.total_reads, axis=0)
    sample_metadata["shannon_otu"] = data_rel.apply(shannon, axis=1)
    sample_metadata["richness_otu"] = (data_rel > 0).sum(axis=1)
    data_abs.to_csv(f"{save_dir}/otu_ab_abs.csv")
    data_rel.to_csv(f"{save_dir}/otu_ab_rel.csv")

    melt = data_rel.melt(ignore_index=False).reset_index()
    melt.columns = ["sample", "otu", "rel_ab"]
    melt = melt.merge(sample_metadata, left_on="sample", right_index=True).merge(
        otu_taxonomy, left_on="otu", right_index=True
    )

    melt_mean = (
        pl.from_pandas(melt)
        .groupby(["otu", tax_level, group_key])
        .agg(rel_ab=pl.col("rel_ab").mean())
    )

    tax_keep = (
        melt_mean.groupby(tax_level)
        .agg(ab=pl.max("rel_ab"))
        .filter(pl.col("ab") > rel_ab_thres)
    )[tax_level]

    tax_table_dict = (
        melt_mean.groupby([tax_level, group_key])
        .agg(rel_ab=pl.col("rel_ab").sum())
        .with_columns(is_good=pl.col(tax_level).is_in(tax_keep))
        .partition_by("is_good", as_dict=True)
    )

    tax_table_good = tax_table_dict[True].drop("is_good")
    try:
        tax_table_bad = (
            tax_table_dict[False]
            .with_columns(pl.lit("others").alias(tax_level))
            .drop("is_good")
        )
    except KeyError:
        tax_table = tax_table_good
    else:
        if keep_others:
            tax_table = pl.concat([tax_table_good, tax_table_bad])
        else:
            tax_table = tax_table_good

    if not keep_unknown:
        tax_table = tax_table.filter(pl.col(tax_level) != "unknown")

    tax_table = (
        tax_table.pivot(
            index=group_key,
            columns=tax_level,
            values="rel_ab",
            aggregate_function="sum",
        )
        .to_pandas()
        .set_index(group_key)
    )

    if not keep_others or not keep_unknown:
        # each row no long adds up to one since low abundance OTUs are filtered
        # genus_table = genus_table.div(genus_table.sum(axis=1), axis=0)
        # family_table = family_table.div(family_table.sum(axis=1), axis=0)
        tax_table = tax_table.div(tax_table.sum(axis=1), axis=0)

    # genus_table.to_csv(f"{save_dir}/genus_rel_ab.csv")
    # family_table.to_csv(f"{save_dir}/family_rel_ab.csv")
    tax_table = tax_table.loc[sample_metadata[group_key]]
    tax_table.to_csv(f"{save_dir}/{tax_level}_rel_ab.csv")

    sample_metadata.to_csv(f"{save_dir}/sample_metadata.csv")



#!/usr/bin/env python3
import argparse
from io import StringIO
import sys
import subprocess
import os
import time
import tempfile
from functools import singledispatch

import numpy as np
import pandas as pd
import polars as pl
from Bio import SeqIO
from Bio.Blast import NCBIWWW, NCBIXML
from Bio import Entrez
from tqdm import tqdm, trange

from masato.utils import print_command


def blast_online(
    input_path: str,
    database: str = "nt",
    batch_size: int = 100,
    email: str = "",
) -> list[str]:
    """
    Submit BLASTn jobs in batches to NCBI for sequences in a given FASTA file and save the results.

    Parameters:
    - input_path: Path to the input FASTA file containing the sequences to query.
    - database: The NCBI database to search against. Default is 'nt'.
    - output_results_path: Path where the combined BLAST results will be saved.
    - batch_size: The number of sequences to include in one query. Default is 100.
    """

    # Read sequences from the FASTA file
    sequences = list(SeqIO.parse(input_path, "fasta"))

    # Prepare batches using list slicing
    batches = [
        sequences[i : i + batch_size] for i in range(0, len(sequences), batch_size)
    ]

    results = []

    # Process each batch
    for batch in tqdm(batches, desc="Submitting BLAST queries"):
        # Convert batch to FASTA format string
        fasta_strings = "\n".join([record.format("fasta") for record in batch])

        result_handle = NCBIWWW.qblast(
            "blastn",
            database,
            fasta_strings,
            format_type="XML",
            megablast=True,
            word_size=64,
            hitlist_size=100,
            filter="mL",
            expect=1e-30,
        )

        # Append result
        results.append(result_handle.read())
    return results


def parse_blast_results(result: str) -> pd.DataFrame:
    results = []
    if not os.path.isdir(result):
        if not result.endswith(".xml"):
            raise ValueError("Result must be a directory containing XML files.")
        with open(result, "r") as f:
            results.append(f.read())
    else:
        for xml_file in os.listdir(result):
            if xml_file.endswith(".xml"):
                with open(os.path.join(result, xml_file), "r") as f:
                    results.append(f.read())
    return pd.concat([extract_info_from_xml(result) for result in results])

    # Write all results to the specified output file
    # with open(output_results_path, "w") as output_file:
    #     for result in results:
    #         output_file.write(result + "\n")

    # print(f"BLAST search completed and results saved to {output_results_path}")


def extract_info_from_xml(xml_string: str, email: str = "") -> pd.DataFrame:
    """
    Extract information from BLAST XML results and include taxonomy IDs.

    Parameters:
    - xml_string: BLAST results in XML format as a string.

    Returns:
    - A pandas DataFrame containing the results with taxonomy IDs.
    """
    records = []

    with StringIO(xml_string.replace("CREATE_VIEW\n\n\n", "")) as xml_handle:
        blast_records = NCBIXML.parse(xml_handle)
        for record in blast_records:
            for alignment in record.alignments:
                for hsp in alignment.hsps:
                    records.append(
                        dict(
                            qseqid=record.query,
                            qlen=record.query_length,
                            sseqid=alignment.hit_id,
                            pident=f"{hsp.identities / float(hsp.align_length) * 100:.2f}",
                            length=hsp.align_length,
                            qstart=hsp.query_start,
                            qend=hsp.query_end,
                            sstart=hsp.sbjct_start,
                            send=hsp.sbjct_end,
                            evalue=hsp.expect,
                            bitscore=hsp.bits,
                            slen=alignment.length,
                            saccession=alignment.accession,
                            sdefinition=alignment.hit_def,
                        )
                    )
    df = pd.DataFrame(records)
    acc2tax = fetch_taxonomy_ids(list(map(str, df["saccession"].unique())), email)
    df["staxids"] = df["saccession"].replace(acc2tax)
    return df


def fetch_taxonomy_ids(
    accession_numbers: list[str], email: str = "", batch_size: int = 100
) -> dict:
    """
    Fetch taxonomy IDs for given accession numbers using NCBI's E-utilities in a single batch request.

    Parameters:
    - accession_numbers: A list of accession numbers.

    Returns:
    - A dictionary mapping accession numbers to taxonomy IDs.
    """
    Entrez.email = email
    taxonomy_ids = {}
    records = []
    # try:
    # Join accession numbers into a single string separated by commas
    for batch_idx, i in enumerate(trange(0, len(accession_numbers), batch_size)):
        successful = False
        while not successful:
            try:
                accession_str = ",".join(accession_numbers[i : i + batch_size])
                handle = Entrez.esummary(db="nucleotide", id=accession_str)
                records.extend(Entrez.read(handle))
                handle.close()
            except RuntimeError as e:
                print(f"An error occurred in batch {batch_idx}: {e}")
                print("Retrying in 5 seconds...")
                time.sleep(5)
            else:
                successful = True

    # Map accession numbers to taxonomy IDs
    for accession, record in zip(accession_numbers, records):
        # The 'Id' field contains the accession number; modify as needed based on the response structure
        tax_id = int(record["TaxId"])
        taxonomy_ids[accession] = tax_id
    # except Exception as e:
    #     print(f"An error occurred while fetching taxonomy IDs: {e}")

    return taxonomy_ids


def blast_local(input_path: str, database: str) -> pd.DataFrame:
    """performs blasts and returns the results"""
    # custom_blast_format = '6 qseqid qlen sseqid pident length qstart qend sstart send evalue bitscore slen staxids'
    custom_blast_format = "6 qseqid qlen sseqid pident length qstart qend sstart send evalue bitscore slen staxids"

    # db_cmd = ["makeblastdb", "-in", subject, "-dbtype", "nucl", "-out", "temp_db"] #command to 1econstruct database
    # sd = subprocess.Popen(db_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE); sd.communicate()
    # temp file
    temp_file = tempfile.NamedTemporaryFile()
    blast_cmd = [
        "blastn",
        "-query",
        input_path,
        "-db",
        database,
        "-num_threads",
        "16",
        "-task",
        "megablast",
        "-word_size",
        "64",  # not too short to speed up the search
        "-dust",
        "yes",
        "-evalue",
        "1e-30",
        "-max_target_seqs",
        "100",
        "-out",
        temp_file.name,
        "-outfmt",
        custom_blast_format,
    ]
    print("Running blastn locally with command:")
    print_command(blast_cmd)
    sp = subprocess.Popen(
        blast_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )  # blast = sp.communicate()
    stdout, stderr = sp.communicate()
    err = stderr.decode() if stderr else ""
    if err:
        print(err)
        print("ERROR: with blast above")
        sys.exit()
    sp.communicate()
    blast_results = pd.read_table(
        temp_file.name,
        header=None,
        names=[
            "qseqid",
            "qlen",
            "sseqid",
            "pident",
            "length",
            "qstart",
            "qend",
            "sstart",
            "send",
            "evalue",
            "bitscore",
            "slen",
            "staxids",
        ],
    )
    return blast_results


def merge_ranges(ranges: list[tuple[int, int]]) -> list[tuple[int, int]]:
    """Merge overlapping ranges and return the merged list."""
    sorted_ranges = sorted(ranges, key=lambda x: x[0])
    merged = []

    for start, end in sorted_ranges:
        if not merged or merged[-1][1] < start:
            merged.append([start, end])
        else:
            merged[-1][1] = max(merged[-1][1], end)

    return merged


def calculate_coverage(ranges: list[tuple[int, int]], qlen: int) -> float:
    """Calculate the exact query coverage based on merged ranges and query length."""
    coverage = sum(end - start + 1 for start, end in ranges) / qlen
    return coverage * 100  # Convert to percentage


@singledispatch
def add_query_coverage(df) -> pd.DataFrame | pl.DataFrame:
    raise NotImplementedError("Unsupported type")


@add_query_coverage.register(pd.DataFrame)
def _(df) -> pd.DataFrame:
    """Add query coverage to the DataFrame without modifying it in-place."""
    # Ensure not modifying the original DataFrame
    df_copy = df.copy()

    # Group by qseqid and aggregate qstart and qend into lists
    agg_df = (
        df_copy.groupby(["qseqid", "sseqid"])
        .agg({"qlen": "first", "qstart": list, "qend": list})
        .reset_index()
    )

    # Calculate merged ranges and coverage for each group
    agg_df["coverage"] = agg_df.apply(
        lambda x: calculate_coverage(
            merge_ranges(list(zip(x["qstart"], x["qend"]))), x["qlen"]
        ),
        axis=1,
    )

    # Merge the coverage back into the original DataFrame
    result_df = pd.merge(
        df_copy,
        agg_df[["qseqid", "sseqid", "coverage"]],
        on=["qseqid", "sseqid"],
        how="left",
    )

    return result_df


def _calc_coverage_pl(struct: pl.Series) -> pl.Series:
    ret = np.zeros(struct["qlen"], dtype=bool)
    for start, end in zip(struct["qstart"], struct["qend"]):
        ret[start - 1 : end] = True
    return ret.sum() / struct["qlen"] * 100


@add_query_coverage.register(pl.DataFrame)
def _(df: pl.DataFrame) -> pl.DataFrame:
    """Implementation for polars DataFrame."""
    return df.join(
        df.group_by(["qseqid", "sseqid"])
        .agg(pl.first("qlen"), pl.col("qstart"), pl.col("qend"))
        .with_columns(
            pl.struct(["qstart", "qend", "qlen"])
            .map_elements(_calc_coverage_pl, return_dtype=pl.Float64)
            .alias("coverage")
        )
        .select(["qseqid", "sseqid", "coverage"]),
        on=["qseqid", "sseqid"],
        how="left",
    )


# if __name__ == "__main__":
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input_",
        help="fasta file of sequences to be classified or a blast xml file",
    )
    parser.add_argument(
        "-o",
        "--blast_output",
        help="output file from blast",
    )
    parser.add_argument(
        "-d", "--database", help="fasta file of sequences to be classified"
    )
    parser.add_argument(
        "-b",
        "--batch_size",
        default=100,
        type=int,
        help="number of sequences to blast at once",
    )
    parser.add_argument("-e", "--email", help="email to use for NCBI E-utilities")

    args = parser.parse_args()

    input_ = args.input_
    database = args.database
    blast_output = args.blast_output
    batch_size = args.batch_size
    email = args.email
    if input_.endswith(".xml") or os.path.isdir(input_):
        df = parse_blast_results(input_)
        df.to_csv(blast_output, index=False, sep="\t")
    elif os.path.isdir(os.path.dirname(database)):
        df = blast_local(input_, database)
        df.to_csv(blast_output, index=False, sep="\t")
    else:
        # must be a fasta file
        results = blast_online(input_, database, batch_size, email)
        blast_result_dir = os.path.join(os.path.dirname(blast_output), "blast_results")
        os.makedirs(blast_result_dir, exist_ok=True)
        for i, result in enumerate(results):
            with open(
                os.path.join(blast_result_dir, f"blast_result_{i}.xml"), "w"
            ) as f:
                print(result, file=f)
        df = parse_blast_results(blast_result_dir)
        df.to_csv(blast_output, index=False, sep="\t")



import subprocess


def map_se(
    reads_fastq: str,
    reference_genome: str,
    *,
    bam_file: str,
    bwa_log: str = "/dev/null",
    mapped_reads: str,
    unmapped_reads: str,
    args: str = "default",
    num_threads: int = 12,
):
    """
    Aligns reads to a reference genome, sorts, indexes the BAM file, and extracts mapped and unmapped reads.

    Args:
        reference_genome (str): Path to the reference genome file.
        reads_fastq (str): Path to the input FASTQ file.
        mapped_reads (str): Path to the output file for mapped reads.
        unmapped_reads (str): Path to the output file for unmapped reads.
        bam_file (str): Path to the sorted BAM file.
        bwa_log (str, optional): Path to the file where stderr of bwa will be logged. Defaults to None.
    """
    if args == "default":
        args = ""
    elif args == "relaxed":
        args = "-k 9 -B 2 -O 4 -E 2 -T 27 "
        # B: mismatch penalty, O: gap open penalty, E: gap extension penalty, T: minimum threshold
    try:
        # Step 2: Align the reads and pipe directly to BAM, then sort
        if not bwa_log:
            bwa_log = "/dev/null"
        align_cmd = (
            f"bwa-mem2 mem -t {num_threads} {args}{reference_genome} {reads_fastq} "
            f"2> {bwa_log} | samtools view -S -b - | samtools sort -o {bam_file}"
        )
        _ = subprocess.run(align_cmd, shell=True, check=True)

        # Step 3: Index the sorted BAM file
        index_cmd = f"samtools index {bam_file}"
        _ = subprocess.run(index_cmd, shell=True, check=True)

        # Step 4: Extract unmapped reads
        extract_mapped_cmd = (
            f"samtools view -F 4 -b {bam_file} | samtools fastq - | pigz > {mapped_reads}"
        )
        _ = subprocess.run(
            extract_mapped_cmd, shell=True, check=True, stderr=subprocess.DEVNULL
        )

        extract_unmapped_cmd = (
            f"samtools view -f 4 -b {bam_file} | samtools fastq - | pigz > {unmapped_reads}"
        )
        _ = subprocess.run(
            extract_unmapped_cmd, shell=True, check=True, stderr=subprocess.DEVNULL
        )

    except subprocess.CalledProcessError as e:
        print(f"An error occurred while processing reads: {e}")


# Example usage:
# process_reads("path/to/ecrec.fna", "path/to/spacers.fq.gz", "path/to/spacers_self.fa", "path/to/spacers_others.fa", "path/to/sorted_aligned_reads.bam", "path/to/bwa_log.txt")
